# -*- coding: utf-8 -*-
"""
Created on Mon May 13 10:48:02 2013

@author: jinpeng

@organization: I2BM, Neurospin, Gif-sur-Yvette, France
@organization: CATI, France
@organization: U{IFR 49<http://www.ifr49.org>}

@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}

"""

if __name__=="__main__":
    import soma_workflow.scheduler
    if soma_workflow.scheduler.DRMAA_LIB_FOUND:
        print "True"
    else:
        print "False"
    



# -----------------------------------------------------------
#  Copyright (C) 2009 StatPro Italia s.r.l.
#
#  StatPro Italia
#  Via G. B. Vico 4
#  I-20123 Milano
#  ITALY
#
#  phone: +39 02 96875 1
#  fax:   +39 02 96875 605
#
#  email: info@riskmap.net
#
#  This program is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the
#  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE. See the license for more details.
# -----------------------------------------------------------
#
#  Author: Enrico Sirola <enrico.sirola@statpro.com>

"""internal helpers"""

import ctypes as _ct
from somadrmaa.wrappers import *
from somadrmaa.errors import error_buffer
import somadrmaa.const as const

_BUFLEN=const.ATTR_BUFFER

try:
    import namedtuple as _nt
except ImportError: # pre 2.6 behaviour
    import nt as _nt

class BoolConverter(object):
    """Helper class to convert to/from bool attributes."""
    def __init__(self, true='y', false='n'):
        self.true=true
        self.false=false
    def to_drmaa(self, value):
        if value: return self.true
        else: return self.false
    def from_drmaa(self, value):
        if value == self.true:
            return True
        else:
            return False

class IntConverter(object):
    """Helper class to convert to/from float attributes."""
    @staticmethod
    def to_drmaa(value):
        return str(value)

    @staticmethod
    def from_drmaa(value):
        return int(value)

class SessionStringAttribute(object):

    def __init__(self, drmaa_function):
        self._f = drmaa_function
    def __get__(self, *args):
        buf = _ct.create_string_buffer(_BUFLEN)
        c(self._f, buf, _ct.sizeof(buf))
        return buf.value

Version = _nt.namedtuple("Version", "major minor")
Version.__str__ = lambda x: "%s.%s" % (x.major, x.minor)
#Version.__doc__ = """\
#An object representing the DRMAA version.
#
#major and minor attributes are int. For DRMAA 1.0, major == 1 and minor == 0.
#"""

class SessionVersionAttribute(object):
    """A Version attribute."""
    def __get__(self, *args):
        major = _ct.c_uint(10)
        minor = _ct.c_uint(10)
        c(drmaa_version, _ct.byref(major), _ct.byref(minor))
        return Version(major.value, minor.value)

class Attribute(object):
    """A DRMAA attribute, to be managed with scalar C DRMAA attribute management functions."""
    def __init__(self, name, type_converter=None):
        """\
Attribute constructor.

:Parameters:
 `name` : string
   name of the attribute to be managed, as seen by the underlying C DRMAA
 `type_converter`
   a converter to translate attribute values to/from the underlying
   implementation. See BoolConverter for an example.
"""
        self.name = name
        self.converter = type_converter
    def __set__(self, instance, value):
        if self.converter:
            v = self.converter.to_drmaa(value)
        else:
            v = value
        c(drmaa_set_attribute, instance, self.name, v)
    def __get__(self, instance, _):
        attr_buffer = create_string_buffer(const.ATTR_BUFFER)
        c(drmaa_get_attribute, instance, self.name, 
          attr_buffer, sizeof(attr_buffer))
        if self.converter:
            return self.converter.from_drmaa(attr_buffer.value)
        else:
            return attr_buffer.value

class VectorAttribute(object):
    """\
A DRMAA attribute representing a list. 

To be managed with vector C DRMAA attribute management functions."""
    def __init__(self, name):
        self.name = name
    def __set__(self, instance, value):
        c(drmaa_set_vector_attribute, instance, self.name, string_vector(value))
    def __get__(self, instance, _):
        return list(vector_attribute_iterator(
                instance, self.name))

class DictAttribute(object):
    """\
A DRMAA attribute representing a python dict.

To be managed with vector C DRMAA attribute management functions."""
    def __init__(self, name):
        self.name = name
    def __set__(self, instance, value):
        v = [ "%s=%s" % (k, v) for (k, v) in value.iteritems() ]
        c(drmaa_set_vector_attribute, instance, self.name, string_vector(v))
    def __get__(self, instance, _):
        x = [ i.split('=', 1) for i in list(vector_attribute_iterator(
                    instance, self.name)) ]
        return dict(x)


def attributes_iterator(attributes):
    try:
        buf = create_string_buffer(const.ATTR_BUFFER)
        while drmaa_get_next_attr_value(attributes, buf, sizeof(buf))\
                != const.NO_MORE_ELEMENTS:
            yield buf.value
    except:
        drmaa_release_attr_values(attributes)
        raise
    else:
        drmaa_release_attr_values(attributes)

def adapt_rusage(rusage):
    "transform a rusage data structure into a dict"
    rv = dict()
    for attr in attributes_iterator(rusage.contents):
        k, v = attr.split('=')
        rv[k] = v
    return rv

def vector_attribute_iterator(jt, attr_name):
    avalues = pointer(POINTER(drmaa_attr_values_t)())
    c(drmaa_get_vector_attribute, jt, attr_name, avalues)
    return attributes_iterator(avalues.contents)

def attribute_names_iterator():
    attrn_p = pointer(POINTER(drmaa_attr_names_t)())
    c(drmaa_get_attribute_names, attrn_p)
    try:
        name = create_string_buffer(_BUFLEN)
        while drmaa_get_next_attr_name(attrn_p.contents, name, _BUFLEN)\
                != const.NO_MORE_ELEMENTS:
            yield name.value
    except:
        drmaa_release_attr_names(attrn_p.contents)
        raise
    else:
        drmaa_release_attr_names(attrn_p.contents)

def vector_attribute_names_iterator():
    attrn_p = pointer(POINTER(drmaa_attr_names_t)())
    c(drmaa_get_vector_attribute_names, attrn_p)
    try:
        name = create_string_buffer(_BUFLEN)
        while drmaa_get_next_attr_name(attrn_p.contents, name, _BUFLEN)\
                != const.NO_MORE_ELEMENTS:
            yield name.value
    except:
        drmaa_release_attr_names(attrn_p.contents)
        raise
    else:
        drmaa_release_attr_names(attrn_p.contents)

def run_bulk_job(jt, start, end, incr=1):
    jids = pointer(POINTER(drmaa_job_ids_t)())
    try:
        c(drmaa_run_bulk_jobs, jids, jt, start, end, incr)
        jid = create_string_buffer(_BUFLEN)
        while drmaa_get_next_job_id(jids.contents, jid, _BUFLEN)\
                != const.NO_MORE_ELEMENTS:
            yield jid.value
    except:
        drmaa_release_job_ids(jids.contents)
        raise
    else:
        drmaa_release_job_ids(jids.contents)

def c(f, *args):
    """An helper function wrapping calls to the C DRMAA functions with error managing code."""
    return f(*(args + (error_buffer, sizeof(error_buffer))))

def string_vector(v):
    vlen = len(v)
    values = (STRING * (vlen + 1))()
    for i, el in enumerate(v):
        values[i] = STRING(el)
    values[vlen] = STRING()
    return values

def attribute_setter(obj, attribute_name):
    "returns a drmaa attribute setter"
    def f(value):
        "setter for %s" % attribute_name
        c(drmaa_set_attribute, obj, attribute_name, value)
    f.__name__ = 'set_'+attribute_name
    return f

def attribute_getter(obj, attribute_name):
    "returns a drmaa attribute setter"
    def f():
        "getter for %s" % attribute_name
        attr_buffer = create_string_buffer(1024)
        c(drmaa_get_attribute, obj, attribute_name, attr_buffer, sizeof(attr_buffer))
        return attr_buffer.value
    f.__name__ = 'get_'+attribute_name
    return f



from __future__ import with_statement

'''
@author: Soizic Laguitton

@organization: I2BM, Neurospin, Gif-sur-Yvette, France
@organization: CATI, France
@organization: U{IFR 49<http://www.ifr49.org>}

@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
'''


import subprocess
import threading
import time
import logging
import os
import sys
import signal
import ctypes
import atexit
import os.path
import socket

import soma_workflow.constants as constants
from soma_workflow.errors import DRMError
from soma_workflow.configuration import LocalSchedulerCfg
from soma_workflow.utils import DetectFindLib

_drmaa_lib_env_name = 'DRMAA_LIBRARY_PATH'

(DRMAA_LIB_FOUND,_lib)=DetectFindLib(_drmaa_lib_env_name,'drmaa')

if DRMAA_LIB_FOUND==True:
    from somadrmaa.errors import *
    from somadrmaa.const import JobControlAction


class Scheduler(object):
  '''
  Allow to submit, kill and get the status of jobs.
  '''
  parallel_job_submission_info = None
  
  logger = None

  is_sleeping = None

  def __init__(self):
    self.parallel_job_submission_info = None
    self.is_sleeping = False

  def sleep(self):
    self.is_sleeping = True

  def wake(self):
    self.is_sleeping = False

  def clean(self):
    pass

  def job_submission(self, job):
    '''
    * job *EngineJob*
    * return: *string*
        Job id for the scheduling system (DRMAA for example)
    '''
    raise Exception("Scheduler is an abstract class!")

  def get_job_status(self, scheduler_job_id):
    '''
    * scheduler_job_id *string*
        Job id for the scheduling system (DRMAA for example)
    * return: *string*
        Job status as defined in constants.JOB_STATUS
    '''
    raise Exception("Scheduler is an abstract class!")

  def get_job_exit_info(self, scheduler_job_id):
    '''
    * scheduler_job_id *string*
        Job id for the scheduling system (DRMAA for example)
    * return: *tuple*
        exit_status, exit_value, term_sig, resource_usage
    '''
    raise Exception("Scheduler is an abstract class!")

  def kill_job(self, scheduler_job_id):
    '''
    * scheduler_job_id *string*
        Job id for the scheduling system (DRMAA for example)
    '''
    raise Exception("Scheduler is an abstract class!")

if DRMAA_LIB_FOUND==True:

  class DrmaaCTypes(Scheduler):
        '''
        Scheduling using a Drmaa session. 
        Contains possible patch depending on the DRMAA impementation. 
        '''
        
        # DRMAA session. DrmaaJobs
        _drmaa = None
        # string
        _drmaa_implementation = None
        # DRMAA doesn't provide an unified way of submitting
        # parallel jobs. The value of parallel_job_submission is cluster dependant. 
        # The keys are:
        #      -Drmaa job template attributes 
        #      -parallel configuration name as defined in soma_workflow.constants
        # dict
        parallel_job_submission_info = None
        
        logger = None
        
        _configured_native_spec = None
  
        tmp_file_path = None
  
        is_sleeping = False
  
          
        def __init__(self, 
                     drmaa_implementation, 
                     parallel_job_submission_info, 
                     tmp_file_path=None, 
                     configured_native_spec=None):
  
          import somadrmaa
  
          self.logger = logging.getLogger('ljp.drmaajs')
          
          self.wake()
   
          self.hostname = socket.gethostname()
      
          self._drmaa_implementation = drmaa_implementation
      
          self.parallel_job_submission_info = parallel_job_submission_info
      
          self._configured_native_spec = configured_native_spec 
      
          self.logger.debug("Parallel job submission info: %s", 
                            repr(parallel_job_submission_info))
  
          if tmp_file_path == None:
             self.tmp_file_path = os.path.abspath("tmp")
          else:
             self.tmp_file_path = os.path.abspath(tmp_file_path) 
       
        def clean(self):
          if self._drmaa_implementation == "PBS":
            tmp_out = os.path.join(self.tmp_file_path, "soma-workflow-empty-job-patch-torque.o")
            tmp_err = os.path.join(self.tmp_file_path, "soma-workflow-empty-job-patch-torque.e")
          
            #print "tmp_out="+tmp_out
            #print "tmp_err="+tmp_err        
  
            if os.path.isfile(tmp_out):
              os.remove(tmp_out)
            if os.path.isfile(tmp_err):
              os.remove(tmp_err) 
  
  
        def close_drmaa_session(self):
          if self._drmaa:
            self._drmaa.exit()
            self._drmaa=None
  
        def __del__(self):
          self.clean()
          self.close_drmaa_session()      
   
        def sleep(self):
          '''
          Some Drmaa sessions expire if they idle too long.  
          '''
          self.close_drmaa_session()
          self.is_sleeping = True
         
  
        def wake(self):
          '''
          Creates a fresh Drmaa session.
          '''
          import somadrmaa
  
          self.is_sleeping = False
          
          if not self._drmaa:
            self._drmaa=somadrmaa.Session()
            self._drmaa.initialize()
 
        def submit_simple_test_job(self, outstr,out_o_file,out_e_file):
          import somadrmaa
          # patch for the PBS-torque DRMAA implementation
          if self._drmaa_implementation == "PBS":
        
                '''
                Create a job to test
                '''
                jobTemplateId = self._drmaa.createJobTemplate()
                jobTemplateId.remoteCommand='echo'
                jobTemplateId.args = ["%s"%(outstr)]
                jobTemplateId.outputPath="%s:%s" %(self.hostname, os.path.join(self.tmp_file_path, "%s"%(out_o_file)))
                jobTemplateId.errorPath="%s:%s" %(self.hostname, os.path.join(self.tmp_file_path, "%s"%(out_e_file)))
    
                #print "jobTemplateId="+repr(jobTemplateId)
                #print "jobTemplateId.remoteCommand="+repr(jobTemplateId.remoteCommand)
                #print "jobTemplateId.args="+repr(jobTemplateId.args)
                #print "jobTemplateId.outputPath="+repr(jobTemplateId.outputPath)
                #print "jobTemplateId.errorPath="+repr(jobTemplateId.errorPath)
  
                jobid=self._drmaa.runJob(jobTemplateId)
                #print "jobid="+jobid
                retval = self._drmaa.wait(jobid, drmaa.Session.TIMEOUT_WAIT_FOREVER) 
                #print "retval="+repr(retval)
                self._drmaa.deleteJobTemplate(jobTemplateId) 
 

 
        def _setDrmaaParallelJob(self, 
                                 drmaa_job_template_id, 
                                 configuration_name, 
                                 max_num_node):
          '''
          Set the DRMAA job template information for a parallel job submission.
          The configuration file must provide the parallel job submission 
          information specific to the cluster in use. 
      
          @type  drmaa_job_template_id: string 
          @param drmaa_job_template_id: id of drmaa job template
          @type  parallel_job_info: tuple (string, int)
          @param parallel_job_info: (configuration_name, max_node_num)
          configuration_name: type of parallel job as defined in soma_workflow.constants 
          (eg MPI, OpenMP...)
          max_node_num: maximum node number the job requests (on a unique machine or 
          separated machine depending on the parallel configuration)
          ''' 
          if self.is_sleeping: self.wake()
          
          self.logger.debug(">> _setDrmaaParallelJob")
          cluster_specific_cfg_name = self.parallel_job_submission_info[configuration_name]
          
          for drmaa_attribute in constants.PARALLEL_DRMAA_ATTRIBUTES:
            value = self.parallel_job_submission_info.get(drmaa_attribute)
            if value: 
              value = value.replace("{config_name}", cluster_specific_cfg_name)
              value = value.replace("{max_node}", repr(max_num_node))
  
  
              setattr(drmaa_job_template_id,drmaa_attribute,value)
  
              self.logger.debug("Parallel job, drmaa attribute = %s, value = %s ",
                                drmaa_attribute, value) 
      
      
          job_env = []
          for parallel_env_v in constants.PARALLEL_JOB_ENV:
            value = self.parallel_job_submission_info.get(parallel_env_v)
            if value: job_env.append((parallel_env_v,value.rstrip()))
  
          drmaa_job_template_id.jobEnvironment=dict(job_env)
  
          self.logger.debug("Parallel job environment : " + repr(job_env)) 
          self.logger.debug("<< _setDrmaaParallelJob")
  
          return drmaa_job_template_id
     
     
        def job_submission(self, job):
          '''
          @type  job: soma_workflow.client.Job
          @param job: job to be submitted
          @rtype: string
          @return: drmaa job id 
          '''
  
          if self.is_sleeping: self.wake()
          # patch for the PBS-torque DRMAA implementation
          command = []
          job_command = job.plain_command()

          ## This is only for the old drmaa version
          ## Now it is not necessary anymore
          #if self._drmaa_implementation == "PBS":
          if False:
            if job_command[0] == 'python':
              job_command[0] = sys.executable
            for command_el in job_command:
              command_el = command_el.replace('"', '\\\"')
              command.append("\"" + command_el + "\"")
            self.logger.debug("PBS case, new command:" + repr(command))
          else:
            command = job_command

       
          self.logger.debug("command: " + repr(command))
          self.logger.debug("job.name=" + repr(job.name))        
  
          stdout_file = job.plain_stdout()
          stderr_file = job.plain_stderr()
          stdin = job.plain_stdin()
       
          try: 
            jobTemplateId = self._drmaa.createJobTemplate()
            jobTemplateId.remoteCommand=command[0]
            jobTemplateId.args = command[1:]
  
            self.logger.info("jobTemplateId="+repr(jobTemplateId)+" command[0]="+repr(command[0])+" command[1:]="+repr(command[1:]))
            self.logger.info("hostname and stdout_file= [%s]:%s" %(self.hostname, stdout_file))
       
            jobTemplateId.outputPath="%s:%s" %(self.hostname, stdout_file)
                      
        
            if job.join_stderrout:
              jobTemplateId.joinFiles="y" 
            else:
              if stderr_file:
                 jobTemplateId.errorPath="%s:%s" %(self.hostname, stderr_file)              
            
            if job.stdin:
              #self.logger.debug("stdin: " + repr(stdin))
              #self._drmaa.setAttribute(drmaaJobId, 
              #                        "drmaa_input_path", 
              #                        "%s:%s" %(self.hostname, stdin))
              self.logger.debug("stdin: " + repr(stdin))
              jobTemplateId.inputPath=stdin
  
       
            working_directory = job.plain_working_directory()
            if working_directory:
              jobTemplateId.workingDirectory=working_directory
  
            self.logger.debug("JOB NATIVE_SPEC " + repr(job.native_specification))
            self.logger.debug("CONFIGURED NATIVE SPEC " + repr(self._configured_native_spec))
            native_spec = None
  
            if job.native_specification:
              native_spec = job.native_specification
            elif self._configured_native_spec:
              native_spec = self._configured_native_spec
            
            if job.queue and native_spec:
              jobTemplateId.nativeSpecification="-q " + str(job.queue) + " " + str(native_spec)
              self.logger.debug("NATIVE specification " + "-q " + str(job.queue) + " " + str(native_spec))
            elif job.queue:
              jobTemplateId.nativeSpecification="-q " + str(job.queue)
              self.logger.debug("NATIVE specification " + "-q " + str(job.queue))
            elif native_spec:
              jobTemplateId.nativeSpecification=str(native_spec)
              self.logger.debug("NATIVE specification " + str(native_spec))
        
            
            if job.parallel_job_info :
              parallel_config_name, max_node_number = job.parallel_job_info
              jobTemplateId=self._setDrmaaParallelJob(jobTemplateId, 
                                        parallel_config_name, 
                                        max_node_number)
             
            if self._drmaa_implementation == "PBS": 
              job_env = []
              for var_name in os.environ.keys():
                #job_env.append(var_name+"="+os.environ[var_name])
                job_env.append((var_name,os.environ[var_name]))
              jobTemplateId.jobEnvironment=dict(job_env)
  
  
            drmaaSubmittedJobId=self._drmaa.runJob(jobTemplateId)
            self._drmaa.deleteJobTemplate(jobTemplateId) 
           
          except DrmaaException, e:
            try:
              f = open(stderr_file, "wa")
              f.write("Error in job submission: %s" %(e))
              f.close()
            except IOError, ioe:
              pass
            self.logger.error("Error in job submission: %s" %(e))
            raise DRMError("Job submission error: %s" %(e))
       
          return drmaaSubmittedJobId      
  
        def kill_job(self, scheduler_job_id):
          if self.is_sleeping: self.wake()
          try:
            self._drmaa.control(scheduler_job_id, JobControlAction.TERMINATE)
          except DrmaaException, e:
            self.logger.critical("%s" %e)
            raise e
  
        def get_job_status(self, scheduler_job_id):
          if self.is_sleeping: self.wake()
          try:
            status = self._drmaa.jobStatus(scheduler_job_id)
          except DrmaaException, e:
            self.logger.error("%s" %(e))
            raise DRMError("%s" %(e))
          return status
  
        def get_job_exit_info(self, scheduler_job_id):
          if self.is_sleeping: self.wake()
  
          res_resourceUsage=[]
          res_status = constants.EXIT_UNDETERMINED
          res_exitValue = 0
          res_termSignal = None        
  
          try:
            self.logger.debug("  ==> Start to find info of job %s"%(scheduler_job_id)) 
            jid_out, exit_value, signaled, term_sig, coredumped, aborted,exit_status,resource_usage=self._drmaa.wait(scheduler_job_id, self._drmaa.TIMEOUT_NO_WAIT)
            
            self.logger.debug("  ==> jid_out="+repr(jid_out))
            self.logger.debug("  ==> exit_value="+repr(exit_value))
            self.logger.debug("  ==> signaled="+repr(signaled))
            self.logger.debug("  ==> term_sig="+repr(term_sig))
            self.logger.debug("  ==> coredumped="+repr(coredumped))
            self.logger.debug("  ==> aborted="+repr(aborted))
            self.logger.debug("  ==> exit_status="+repr(exit_status))
            self.logger.debug("  ==> resource_usage="+repr(resource_usage))
    
            if aborted == 1:
              res_status=constants.EXIT_ABORTED
            else:
              if exit_value == 1:
                res_status=constants.FINISHED_REGULARLY
                res_exitValue=exit_status
              else :
                if signaled == 1:
                  res_status=constants.FINISHED_TERM_SIG
                  res_termSignal=term_sig
                else:
                  res_status=constants.FINISHED_UNCLEAR_CONDITIONS
    
  
            res_resourceUsage = ''
            for k,v in resource_usage.iteritems():
              res_resourceUsage = res_resourceUsage + k + '=' + v + ' '
  
  
          except ExitTimeoutException:
            res_status = constants.EXIT_UNDETERMINED
            self.logger.debug("  ==> self._drmaa.wait time out")
  
          return (res_status,res_exitValue , res_termSignal, res_resourceUsage)
else:

  class DrmaaCTypes(Scheduler):pass


class LocalScheduler(Scheduler):
  '''
  Allow to submit, kill and get the status of jobs.
  Run on one machine without dependencies.

  * _proc_nb *int*

  * _queue *list of scheduler jobs ids*
  
  * _jobs *dictionary job_id -> soma_workflow.engine_types.EngineJob*
  
  * _processes *dictionary job_id -> subprocess.Popen*

  * _status *dictionary job_id -> job status as defined in constants*

  * _exit_info * dictionay job_id -> exit info*

  * _loop *thread*

  * _interval *int*

  * _look *threading.RLock*
  '''
  parallel_job_submission_info = None
  
  logger = None

  _proc_nb = None

  _queue = None

  _jobs = None 
  
  _processes = None

  _status = None

  _exit_info = None

  _loop = None

  _interval = None

  _lock = None

  def __init__(self, proc_nb=1, interval=1):
    super(LocalScheduler, self).__init__()
  
    self.parallel_job_submission_info = None

    self._proc_nb = proc_nb
    self._interval = interval
    self._queue = []
    self._jobs = {}
    self._processes = {}
    self._status = {}
    self._exit_info = {}

    self._lock = threading.RLock()

    self.stop_thread_loop = False

    def loop(self):
      while not self.stop_thread_loop:
        with self._lock:
          self._iterate()
        time.sleep(self._interval)

    self._loop = threading.Thread(name="scheduler_loop",
                                  target=loop,
                                  args=[self])
    self._loop.setDaemon(True)
    self._loop.start()

    atexit.register(LocalScheduler.end_scheduler_thread, self)

  def change_proc_nb(self, proc_nb):
    with self._lock:
      self._proc_nb = proc_nb

  def change_interval(self, interval):
    with self._lock:
      self._interval = interval

  def end_scheduler_thread(self):
    with self._lock:
      self.stop_thread_loop = True
      self._loop.join()
      #print "Soma scheduler thread ended nicely."

  def _iterate(self):
    # Nothing to do if the queue is empty and nothing is running
    if not self._queue and not self._processes:
      return
    #print "#############################"
    # Control the running jobs
    ended_jobs = []
    for job_id, process in self._processes.iteritems():
      ret_value = process.poll()
      #print "job_id " + repr(job_id) + " ret_value " + repr(ret_value)
      if ret_value != None:
        ended_jobs.append(job_id) 
        self._exit_info[job_id] = (constants.FINISHED_REGULARLY,
                                     ret_value,
                                     None,
                                     None)

    # update for the ended job
    for job_id in ended_jobs:
      #print "updated job_id " + repr(job_id) + " status DONE"
      self._status[job_id] = constants.DONE
      del self._processes[job_id]

    # run new jobs
    while (self._queue and
           len(self._processes) < self._proc_nb):
      job_id = self._queue.pop(0)
      job = self._jobs[job_id]
      #print "new job " + repr(job.job_id)
      process = LocalScheduler.create_process(job)
      if process == None:
        self._exit_info[job.job_id] = (constants.EXIT_ABORTED,
                                   None,
                                   None,
                                   None)
        self._status[job.job_id] = constants.FAILED
      else:
        self._processes[job.job_id] = process
        self._status[job.job_id] = constants.RUNNING


  @staticmethod
  def create_process(engine_job):
    '''
    * engine_job *EngineJob*
   
    * returns: *Subprocess process* 
    '''

    command = engine_job.plain_command()

    stdout = engine_job.plain_stdout()
    stdout_file = None
    if stdout:
      try:
        stdout_file = open(stdout, "wb")
      except Exception, e:
        return None

    stderr = engine_job.plain_stderr()
    stderr_file = None
    if stderr:
      try:
        stderr_file = open(stderr, "wb")
      except Exception, e:
        return None

    stdin = engine_job.plain_stdin()
    stdin_file = None
    if stdin:
      try:
        stdin_file = open(stdin, "rb")
      except Exception, e:
        if stderr:
          stderr_file = open(stderr, "wb")
          s = '%s: %s \n' %(type(e), e)
          stderr_file.write(s)
          stderr_file.close()
        else:
          stdout_file = open(stdout, "wb")
          s = '%s: %s \n' %(type(e), e)
          stdout_file.write(s)
          stdout_file.close()
        return None

    working_directory = engine_job.plain_working_directory()
    
    try:
      process = subprocess.Popen( command,
                                  stdin=stdin_file,
                                  stdout=stdout_file,
                                  stderr=stderr_file,
                                  cwd=working_directory)

    
    except Exception, e:
      if stderr:
        stderr_file = open(stderr, "wb")
        s = '%s: %s \n' %(type(e), e)
        stderr_file.write(s)
        stderr_file.close()
      else:
        stdout_file = open(stdout, "wb")
        s = '%s: %s \n' %(type(e), e)
        stdout_file.write(s)
        stdout_file.close()
      return None

    return process


  def job_submission(self, job):
    '''
    * job *EngineJob*
    * return: *string*
        Job id for the scheduling system (DRMAA for example)
    '''
    if not job.job_id or job.job_id == -1:
      raise LocalSchedulerError("Invalid job: no id")
    with self._lock:
      #print "job submission " + repr(job.job_id)
      self._queue.append(job.job_id)
      self._jobs[job.job_id] = job
      self._status[job.job_id] = constants.QUEUED_ACTIVE
      self._queue.sort(key=lambda job_id: self._jobs[job_id].priority, 
                       reverse=True)
    return job.job_id


  def get_job_status(self, scheduler_job_id):
    '''
    * scheduler_job_id *string*
        Job id for the scheduling system (DRMAA for example)
    * return: *string*
        Job status as defined in constants.JOB_STATUS
    '''
    if not scheduler_job_id in self._status:
      raise LocalSchedulerError("Unknown job.")
   
    status = self._status[scheduler_job_id]
    return status

  def get_job_exit_info(self, scheduler_job_id):
    '''
    * scheduler_job_id *string*
        Job id for the scheduling system (DRMAA for example)
    * return: *tuple*
        exit_status, exit_value, term_sig, resource_usage
    '''
    # TBI errors
    with self._lock:
      exit_info = self._exit_info[scheduler_job_id]
      del self._exit_info[scheduler_job_id]
    return exit_info


  def kill_job(self, scheduler_job_id):
    '''
    * scheduler_job_id *string*
        Job id for the scheduling system (DRMAA for example)
    '''
    # TBI Errors
    
    with self._lock:
      #print "kill job " + repr(scheduler_job_id)
      if scheduler_job_id in self._processes:
        #print "    => kill the process "
        process = self._processes[scheduler_job_id]
        if sys.version_info < (2, 6):
          if sys.platform == 'win32':
            PROCESS_TERMINATE = 1
            handle = ctypes.windll.kernel32.OpenProcess(PROCESS_TERMINATE, 
                                               False, 
                                               process.pid)
            ctypes.windll.kernel32.TerminateProcess(handle, -1)
            ctypes.windll.kernel32.CloseHandle(handle)
          else:
            os.kill(process.pid, signal.SIGKILL)
            os.wait()
        else:
          process.kill()

        del self._processes[scheduler_job_id]
        self._status[scheduler_job_id] = constants.FAILED
        self._exit_info[scheduler_job_id] = (constants.USER_KILLED,
                                              None,
                                              None,
                                              None)
      elif scheduler_job_id in self._queue:
        #print "    => removed from queue "
        self._queue.remove(scheduler_job_id)
        del self._jobs[scheduler_job_id]
        self._status[scheduler_job_id] = constants.FAILED
        self._exit_info[scheduler_job_id] = (constants.EXIT_ABORTED,
                                              None,
                                              None,
                                              None)


class ConfiguredLocalScheduler(LocalScheduler):
  '''
  Local scheduler synchronized with a configuration object.
  '''

  _config = None

  def __init__(self, config):
    '''
    * config *LocalSchedulerCfg*
    '''
    super(ConfiguredLocalScheduler, self).__init__(config.get_proc_nb(), 
                                           config.get_interval())
    self._config = config

    self._config.addObserver(self,
                             "update_from_config", 
                             [LocalSchedulerCfg.PROC_NB_CHANGED, LocalSchedulerCfg.INTERVAL_CHANGED])


  def update_from_config(self, observable, event, msg):
    if event == LocalSchedulerCfg.PROC_NB_CHANGED:
      self.change_proc_nb(self._config.get_proc_nb())
    if event == LocalSchedulerCfg.PROC_NB_CHANGED:
      self.change_interval(self._config.get_interval())
    self._config.save_to_file()
    




# -*- coding: utf-8 -*-

'''
@author: Soizic Laguitton

@organization: I2BM, Neurospin, Gif-sur-Yvette, France
@organization: CATI, France
@organization: U{IFR 49<http://www.ifr49.org>}

@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
'''


#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------

import soma_workflow.constants as constants

#-------------------------------------------------------------------------------
# Classes and functions
#-------------------------------------------------------------------------------


class Job(object):
  '''
  Job representation.

  .. note::
    The command is the only argument required to create a Job.
    It is also useful to fill the job name for the workflow display in the GUI.

  **command**: *sequence of string or/and FileTransfer or/and SharedResourcePath or/and tuple (FileTransfer, relative_path) or/and sequence of FileTransfer or/and sequence of SharedResourcePath or/and sequence of tuple (FileTransfer, relative_path)*

    The command to execute. It can not be empty. In case of a shared file system
    the command is a sequence of string.

    In the other cases, the FileTransfer and SharedResourcePath objects will be
    replaced by the appropriate path before the job execution.

    The tuples (FileTransfer, relative_path) can be used to refer to a file in a
    transfered directory.

    The sequences of FileTransfer, SharedResourcePath or tuple (FileTransfer,
    relative_path) will be replaced by the string "['path1', 'path2', 'path3']"
    before the job execution. The FileTransfer, SharedResourcePath or tuple
    (FileTransfer, relative_path) are replaced by the appropriate path inside
    the sequence.

  **name**: *string*
    Name of the Job which will be displayed in the GUI

  **referenced_input_files**: *sequence of FileTransfer*
    List of the FileTransfer which are input of the Job. In other words,
    FileTransfer which are requiered by the Job to run. It includes the
    stdin if you use one.

  **referenced_output_files**: *sequence of FileTransfer*
    List of the FileTransfer which are output of the Job. In other words, the
    FileTransfer which will be created or modified by the Job.

  **stdin**: *string or FileTransfer or SharedRessourcePath*
    Path to the file which will be read as input stream by the Job.

  **join_stderrout**: *boolean*
    Specifies whether the error stream should be mixed with the output stream.

  **stdout_file**: *string or FileTransfer or SharedRessourcePath*
    Path of the file where the standard output stream of the job will be
    redirected.

  **stderr_file**: *string or FileTransfer or SharedRessourcePath*
    Path of the file where the standard error stream of the job will be
    redirected.

  .. note::
    Set stdout_file and stderr_file only if you need to redirect the standard
    output to a specific file. Indeed, even if they are not set the standard
    outputs will always be available through the WorklfowController API.

  **working_directory**: *string or FileTransfer or SharedRessourcePath*
    Path of the directory where the job will be executed. The working directory
    is useful if your Job uses relative file path for example.

  **priority**: *int*
    Job priority: 0 = low priority. If several Jobs are ready to run at the 
    same time the jobs with higher priority will be submitted first.

  **native_specification**: *string*
    Some specific option/function of the computing resource you want to use 
    might not be available among the list of Soma-workflow Job attributes.
    Use the native specification attribute to use these specific functionalities. 
    If a native_specification is defined here, the configured native 
    specification will be ignored (documentation configuration item: NATIVE_SPECIFICATION).

    *Example:* Specification of a job walltime and more:
      * using a PBS cluster: native_specification="-l walltime=10:00:00,pmem=16gb" 
      * using a SGE cluster: native_specification="-l h_rt=10:00:00"

  **parallel_job_info**: *tuple(string, int)*
    The parallel job information must be set if the Job is parallel (ie. made to
    run on several CPU).
    The parallel job information is a tuple: (name of the configuration,
    maximum number of CPU used by the Job).
    The configuration name is the type of parallel Job. Example: MPI or OpenMP.

    .. warning::
      The computing resources must be configured explicitly to use this feature.

  **user_storage**: *picklable object*
    For the user needs, any small and picklable object can be stored here.

  ..
    **disposal_time_out**: int
    Only requiered outside of a workflow
  '''

  # sequence of sequence of string or/and FileTransfer or/and SharedResourcePath or/and tuple (relative_path, FileTransfer) or/and sequence of FileTransfer or/and sequence of SharedResourcePath or/and sequence of tuple (relative_path, FileTransfers.)
  command = None

  # string
  name = None

  # sequence of FileTransfer
  referenced_input_files = None

  # sequence of FileTransfer
  referenced_output_files = None

  # string (path)
  stdin = None

  # boolean
  join_stderrout = None

  # string (path)
  stdout_file = None

  # string (path)
  stderr_file = None

  # string (path)
  working_directory = None

  # int 
  priority = None

  # string
  native_specification = None

  # tuple(string, int)
  parallel_job_info = None

  # int (in hours)
  disposal_timeout = None

  # any small and picklable object needed by the user
  user_storage = None

  def __init__( self,
                command,
                referenced_input_files=None,
                referenced_output_files=None,
                stdin=None,
                join_stderrout=False,
                disposal_timeout=168,
                name=None,
                stdout_file=None,
                stderr_file=None,
                working_directory=None,
                parallel_job_info=None,
                priority=0,
                native_specification=None,
                user_storage=None):
    if not name:
      self.name = command[0]
    else:
      self.name = name
    self.command = command
    if referenced_input_files:
      self.referenced_input_files = referenced_input_files
    else: self.referenced_input_files = []
    if referenced_output_files:
      self.referenced_output_files = referenced_output_files
    else: self.referenced_output_files = []
    self.stdin = stdin
    self.join_stderrout = join_stderrout
    self.disposal_timeout = disposal_timeout
    self.stdout_file = stdout_file
    self.stderr_file = stderr_file
    self.working_directory = working_directory
    self.parallel_job_info = parallel_job_info
    self.priority = priority
    self.native_specification = native_specification

  
  def _attributs_equal(self, el_list, other_el_list):
    if not len(el_list) == len(other_el_list):
      return False
    for i in range(0, len(el_list)):
      if isinstance(el_list[i], FileTransfer) or\
         isinstance(el_list[i], SharedResourcePath):
        if not el_list[i].attributs_equal(other_el_list[i]):
          return False
      elif isinstance(el_list[i], tuple):
        if not isinstance(other_el_list[i], tuple) or\
           not len(el_list[i]) == len(other_el_list[i]):
          return False
        if not el_list[i][0].attributs_equal(other_el_list[i][0]):
          return False
        if not el_list[i][1] == other_el_list[i][1]:
          return False
      elif isinstance(el_list[i], list):
        if not isinstance(other_el_list[i], list):
          return False
        if not self._attributs_equal(el_list[i], other_el_list[i]):
          return False
      elif not el_list[i] == other_el_list[i]:
        return False
    return True

  def attributs_equal(self, other):
    # TODO a better solution would be to overload __eq__ and __neq__ operator
    # however these operators are used in soma-workflow to test if 
    # two objects are the same instancete. These tests have to be replaced 
    # first using the id python function.
    if not isinstance(other, self.__class__):
      return False
    seq_attributs = [
                     "command", 
                     "referenced_input_files",
                     "referenced_output_files"]
    for attr_name in seq_attributs:
      attr = getattr(self, attr_name)
      other_attr = getattr(other, attr_name)
      if not self._attributs_equal(attr, other_attr):
        return False

    attributs = [
                 "name", 
                 "stdin",
                 "join_stderrout",
                 "stdout_file",
                 "stderr_file",
                 "working_directory",
                 "priority",
                 "native_specification",
                 "parallel_job_info",
                 "disposal_timeout",
                 ]
    for attr_name in attributs:
      attr = getattr(self, attr_name)
      other_attr = getattr(other, attr_name)
      if isinstance(attr, FileTransfer) or\
         isinstance(attr, SharedResourcePath):
        if not attr.attributs_equal(other_attr):
          return False
      elif not attr == other_attr:
        return False
    return True


  @classmethod
  def from_dict(cls,
                d, 
                tr_from_ids,
                srp_from_ids):
    '''
     * d *dictionary*
     * tr_from_id *id -> FileTransfer*
     * srp_from_id *id -> SharedResourcePath*
    '''
    job = cls(command=d["command"])
    for key, value in d.iteritems():
      setattr(job, key, value)
  
    new_command = list_from_serializable(job.command,
                                       tr_from_ids,
                                       srp_from_ids)
    job.command = new_command
  
    if job.referenced_input_files:
      ref_in_files = list_from_serializable(job.referenced_input_files,
                                          tr_from_ids,
                                          srp_from_ids)
      job.referenced_input_files = ref_in_files
  
    if job.referenced_output_files:
      ref_out_files = list_from_serializable(job.referenced_output_files,
                                           tr_from_ids,
                                           srp_from_ids)
      job.referenced_output_files = ref_out_files
  
    if job.stdin:
      job.stdin = from_serializable(job.stdin,
                                    tr_from_ids,
                                    srp_from_ids)
    if job.stdout_file:
      job.stdout_file = from_serializable(job.stdout_file,
                                          tr_from_ids,
                                          srp_from_ids)
    if job.stderr_file:
      job.stderr_file = from_serializable(job.stderr_file,
                                          tr_from_ids,
                                          srp_from_ids)
    if job.working_directory:  
      job.working_directory = from_serializable(job.working_directory,
                                                tr_from_ids,
                                                srp_from_ids)
  
    return job
  
  def to_dict(self,
              id_generator, 
              transfer_ids, 
              shared_res_path_id):
    '''
    * id_generator *IdGenerator*
    * transfer_ids *dict: client.FileTransfer -> int*
        This dictonary will be modified.
    * shared_res_path_id *dict: client.SharedResourcePath -> int*
        This dictonary will be modified.
    '''
    job_dict = {}
  
    attributs = [
                 "name", 
                 "join_stderrout",
                 "priority",
                 "native_specification",
                 "parallel_job_info",
                 "disposal_timeout",
                ]
  
    for attr_name in attributs:
      job_dict[attr_name] = getattr(self, attr_name)
  
    # command, referenced_input_files, referenced_output_files   
    # stdin, stdout_file, stderr_file and working_directory
    # can contain FileTransfer et SharedResourcePath.
   
    ser_command = list_to_serializable(self.command, 
                                       id_generator, 
                                       transfer_ids, 
                                       shared_res_path_id)
  
  
    job_dict['command'] = ser_command  
  
    if self.referenced_input_files:
      ser_ref_in_files = list_to_serializable(self.referenced_input_files, 
                                              id_generator, 
                                              transfer_ids, 
                                              shared_res_path_id)
      job_dict['referenced_input_files'] = ser_ref_in_files
  
    if self.referenced_output_files:
      ser_ref_out_files = list_to_serializable(self.referenced_output_files, 
                                               id_generator, 
                                               transfer_ids, 
                                               shared_res_path_id)
      job_dict['referenced_output_files'] = ser_ref_out_files
  
  
    if self.stdin:
      job_dict['stdin'] = to_serializable(self.stdin,
                                          id_generator, 
                                          transfer_ids, 
                                          shared_res_path_id)
  
    if self.stdout_file:
      job_dict['stdout_file'] = to_serializable(self.stdout_file,
                                                id_generator, 
                                                transfer_ids, 
                                                shared_res_path_id)
  
  
    if self.stderr_file:
      job_dict['stderr_file'] = to_serializable(self.stderr_file,
                                                id_generator, 
                                                transfer_ids, 
                                                shared_res_path_id)
  
  
    if self.working_directory:  
      job_dict['working_directory'] = to_serializable(self.working_directory,
                                                      id_generator, 
                                                      transfer_ids, 
                                                      shared_res_path_id)
  
  
    return job_dict




class Workflow(object):
  '''
  Workflow representation.

  **name**: *string*

  **jobs**: *sequence of Job*
    Workflow jobs.

  **dependencies**: *sequence of tuple (Job, Job)*
    Dependencies between the jobs of the workflow.
    If a job_a needs to be executed before a job_b can run: the tuple
    (job_a, job_b) must be added to the workflow dependencies. job_a and job_b
    must belong to workflow.jobs.

  **root_group**: *sequence of Job and/or Group*
    Recursive description of the workflow hierarchical structure. For displaying
    purpose only.

  .. note::
    root_group is only used to display nicely the workflow in the GUI. It
    does not have any impact on the workflow execution.

    If root_group is not set, all the jobs of the workflow will be
    displayed at the root level in the GUI tree view.

  **user_storage**: *picklable object*
    For the user needs, any small and picklable object can be stored here.

  **name**: *string*
    Name of the workflow which will be displayed in the GUI.
    Default: workflow_id once submitted

  '''
  # string
  name = None

  # sequence of Job
  jobs = None

  # sequence of tuple (Job, Job)
  dependencies = None

  # sequence of Job and/or Group
  root_group = None

  # sequence of Groups built from the root_group
  groups = None

  # any small and picklable object needed by the user
  user_storage = None

  def __init__(self,
               jobs,
               dependencies=None,
               root_group=None,
               disposal_timeout=168,
               user_storage=None,
               name=None):

    self.name = name
    self.jobs = jobs
    if dependencies != None:
      self.dependencies = dependencies
    else:
      self.dependencies = []
    self.disposal_timeout = 168

    # Groups
    if root_group:
      if isinstance(root_group, Group):
        self.root_group = root_group.elements
      else:
        self.root_group = root_group
      self.groups = []
      to_explore = []
      for element in self.root_group:
        if isinstance(element, Group):
          to_explore.append(element)
      while to_explore:
        group = to_explore.pop()
        self.groups.append(group)
        for element in group.elements:
          if isinstance(element, Group):
            to_explore.append(element)
    else:
      self.root_group = self.jobs
      self.groups = []

  def attributs_equal(self, other):
    if not isinstance(other, self.__class__):
      return False
    seq_attributs = [
                     "jobs",
                     "dependencies",
                     "root_group",
                     "groups"
                     ]
    for attr_name in seq_attributs:
      attr = getattr(self, attr_name)
      other_attr = getattr(other, attr_name)
      if not len(attr) == len(other_attr):
        return False
      for i in range(0, len(attr)):
        if isinstance(attr[i], Job) or\
         isinstance(attr[i], Group):
          if not attr[i].attributs_equal(other_attr[i]):
            return False
        elif isinstance(attr[i], tuple):
          if not isinstance(other_attr[i], tuple) or\
             not len(attr[i]) == len(other_attr[i]):
            return False
          if not attr[i][0].attributs_equal(other_attr[i][0]):
            return False
          if not attr[i][1].attributs_equal(other_attr[i][1]):
            return False
        elif not attr[i] == other_attr[i]:
          return False
    return self.name == other.name

  def to_dict(self):
    '''
    The keys must be string to serialize with JSON.
    '''
    #TODO user_storage
    id_generator = IdGenerator()
    job_ids = {} # Job -> id
     
    wf_dict = {}
  
    wf_dict["name"] = self.name
   
    new_jobs = []
    for job in self.jobs:
      ident = id_generator.generate_id()
      new_jobs.append(ident)
      job_ids[job] = ident
    wf_dict["jobs"] = new_jobs
    
    new_dependencies = []
    for dep in self.dependencies:
      if dep[0] not in job_ids or dep[1] not in job_ids:
        raise Exception("Unknown jobs in dependencies.")
      new_dependencies.append((job_ids[dep[0]], job_ids[dep[1]]))
    wf_dict["dependencies"] = new_dependencies
  
    group_ids = {}
    new_groups = []
    for group in self.groups:
      ident = id_generator.generate_id()
      new_groups.append(ident)
      group_ids[group] = ident
    wf_dict["groups"] = new_groups
  
    new_root_group = []
    for element in self.root_group:
      if element in job_ids:
        new_root_group.append(job_ids[element])
      elif element in group_ids:
        new_root_group.append(group_ids[element])
      else:
        raise Exception("Unknown root group element.") 
    wf_dict["root_group"] = new_root_group
  
    ser_groups = {}
    for group, group_id in group_ids.iteritems():
     ser_groups[str(group_id)] = group.to_dict(group_ids, job_ids)
    wf_dict["serialized_groups"] = ser_groups
  
    ser_jobs = {}
    transfer_ids = {} # FileTransfer -> id
    shared_res_path_ids = {} #SharedResourcePath -> id
    for job, job_id in job_ids.iteritems():
      ser_jobs[str(job_id)] = job.to_dict(id_generator,
                                          transfer_ids,
                                          shared_res_path_ids)
    wf_dict["serialized_jobs"] = ser_jobs
  
    ser_transfers = {}
    for file_transfer, transfer_id in transfer_ids.iteritems():
      ser_transfers[str(transfer_id)] = file_transfer.to_dict()
    wf_dict["serialized_file_transfers"] = ser_transfers
  
    ser_srp = {}
    for srp, srp_id in shared_res_path_ids.iteritems():
      ser_srp[str(srp_id)] = srp.to_dict()
    wf_dict["serialized_shared_res_paths"] = ser_srp
  
    return wf_dict 
  
  @classmethod
  def from_dict(cls, d):
    name = d["name"]
    
    #shared resource paths
    serialized_srp = d["serialized_shared_res_paths"]
    srp_from_ids = {}
    for srp_id, srp_d in serialized_srp.iteritems():
      srp = SharedResourcePath.from_dict(srp_d)
      srp_from_ids[int(srp_id)] = srp
  
    #file transfers
    serialized_tr = d["serialized_file_transfers"]
    tr_from_ids = {}
    for tr_id, tr_d in serialized_tr.iteritems():
      file_transfer = FileTransfer.from_dict(tr_d)
      tr_from_ids[int(tr_id)] = file_transfer
    
    #jobs
    serialized_jobs = d["serialized_jobs"]
    job_from_ids = {}
    for job_id, job_d in serialized_jobs.iteritems():
      job = Job.from_dict(job_d, tr_from_ids, srp_from_ids)
      job_from_ids[int(job_id)] = job
    jobs = job_from_ids.values()
  
    #groups
    serialized_groups = d["serialized_groups"]
    group_from_ids = {}
    to_convert = serialized_groups.keys()
    converted_or_stuck = False
    while not converted_or_stuck  :
      new_converted = []
      for group_id in to_convert: 
        group = Group.from_dict(serialized_groups[group_id],
                                group_from_ids,
                                job_from_ids)
        if group != None:
          new_converted.append(group_id)
          group_from_ids[int(group_id)] = group
      for group_id in new_converted: to_convert.remove(group_id)
      converted_or_stuck = not to_convert or not new_converted
    groups = group_from_ids.values()
            
    #root group
    id_root_group = d["root_group"] 
    root_group = []
    for el_id in id_root_group:
      if el_id in group_from_ids:
        root_group.append(group_from_ids[el_id])
      elif el_id in job_from_ids:
        root_group.append(job_from_ids[el_id])
  
    #dependencies
    dependencies = []
    id_dependencies = d["dependencies"]
    for id_dep in id_dependencies:
      dep = (job_from_ids[id_dep[0]], job_from_ids[id_dep[1]])
      dependencies.append(dep)
  
    workflow = cls(jobs, 
                   dependencies, 
                   root_group=root_group,
                   user_storage=None,
                   name=name)  
  
    return workflow
   
  




class Group(object):
  '''
  Hierarchical structure of a workflow.

  .. note:
    It only has a displaying role and does not have any impact on the workflow
    execution.

  **elements**: *sequence of Job and/or Group*
    The elements (Job or Group) belonging to the group.
 
  **name**: *string*
    Name of the Group which will be displayed in the GUI.
 
  **user_storage**: *picklable object*
    For the user needs, any small and picklable object can be stored here.
  '''
  #string
  name = None

  #sequence of Job and/or Group
  elements = None

  # any small and picklable object needed by the user
  user_storage = None


  def __init__(self, elements, name, user_storage=None):
    '''
    @type  elements: sequence of Job and/or Group
    @param elements: the elements belonging to the group
    @type  name: string
    @param name: name of the group
    '''
    self.elements = elements
    self.name = name

  def attributs_equal(self, other):
    if not isinstance(other, self.__class__):
      return False
    if len(self.elements) != len(other.elements):
      return False
    for i in range(0, len(self.elements)):
      if not self.elements[i].attributs_equal(other.elements[i]):
        return False
    if self.name != other.name:
      return False
    return True

  def to_dict(self, group_ids, job_ids):
    group_dict = {}
    group_dict["name"] = self.name
    new_gp_elements = []
    for element in self.elements:
      if element in job_ids:
        new_gp_elements.append(job_ids[element])
      elif element in group_ids:
        new_gp_elements.append(group_ids[element])
      else:
        raise Exception("Unknown group element.")
    group_dict["elements"] = new_gp_elements
    return group_dict
  
  @classmethod
  def from_dict(cls, d, group_from_ids, job_from_ids):
    id_elements = d["elements"]
    elements = []
    for el_id in id_elements:
      if el_id in job_from_ids:
        elements.append(job_from_ids[el_id])
      elif el_id in group_from_ids:
        elements.append(group_from_ids[el_id])
      else:
        return None
    
    name = d["name"]
    group = cls(elements, name)
    return group
  


class FileTransfer(object):
  '''
  File/directory transfer representation

  .. note::
    FileTransfers objects are only required if the user and computing resources
    have a separate file system.

  **client_path**: *string*
    Path of the file or directory on the user's file system.

  **initial_status**: *constants.FILES_DO_NOT_EXIST or constants.FILES_ON_CLIENT*
    * constants.FILES_ON_CLIENT for workflow input files
      The file(s) will need to be transfered on the computing resource side
    * constants.FILES_DO_NOT_EXIST for workflow output files
      The file(s) will be created by a job on the computing resource side.

  **client_paths**: *sequence of string*
    Sequence of path. Files to transfer if the FileTransfers concerns a file
    series or if the file format involves several associated files or
    directories (see the note below).

  **name**: *string*
    Name of the FileTransfer which will be displayed in the GUI.
    Default: client_path + "transfer"

  .. note::
    Use client_paths if the transfer involves several associated files and/or
    directories. Examples:

      * file series
      * file format associating several file and/or directories
        (ex: a SPM images are stored in 2 associated files: .img and .hdr)
        In this case, set client_path to one the files (ex: .img) and
        client_paths contains all the files (ex: .img and .hdr files)

     In other cases (1 file or 1 directory) the client_paths must be set to None.
  '''

  # string
  client_path = None

  # sequence of string
  client_paths = None

  # constants.FILES_DO_NOT_EXIST constants.FILES_ON_CLIENT
  initial_status = None

  # int (hours)
  disposal_timeout = None

  # string
  name = None

  def __init__( self,
                is_input,
                client_path,
                disposal_timeout = 168,
                name = None,
                client_paths = None,
                ):
    if name:
      ft_name = name
    else:
      ft_name = client_path + "transfer"
    self.name = ft_name

    self.client_path = client_path
    self.disposal_timeout = disposal_timeout

    self.client_paths = client_paths

    if is_input:
      self.initial_status = constants.FILES_ON_CLIENT
    else:
      self.initial_status = constants.FILES_DO_NOT_EXIST

  def attributs_equal(self, other):
    if not isinstance(other, self.__class__):
      return False
    attributs = [
                "client_path",
                "client_paths",
                "initial_status",
                "disposal_timeout",
                "name",
                ]
    for attr_name in attributs:
      attr = getattr(self, attr_name)
      other_attr = getattr(other, attr_name)
      if not other_attr == attr:
        return False
    return True

  def to_dict(self):
    transfer_dict = {}
    attributs = [
                 "client_path",
                 "client_paths",
                 "initial_status",
                 "disposal_timeout",
                 "name",
                ]
    for attr_name in attributs:
      transfer_dict[attr_name] = getattr(self, attr_name)
  
    return transfer_dict
  
  @classmethod
  def from_dict(cls, d):
    transfer = cls(is_input=True, 
                   client_path="foo")
    for key, value in d.iteritems():
      setattr(transfer, key, value)
    return transfer
  




class SharedResourcePath(object):
  '''
  Representation of path which is valid on either user's or computing resource
  file system.

  .. note::
    SharedResourcePath objects are only required if the user and computing
    resources have a separate file system.

  **namespace**: *string*
    Namespace for the path. That way several applications can use the same
    identifiers without risk.

  **uuid**: *string*
    Identifier of the absolute path.

  **relative_path**: *string*
    Relative path of the file if the absolute path is a directory path.

  .. warning::
    The namespace and uuid must exist in the translations files configured on
    the computing resource side.
  '''

  relative_path = None

  namespace = None

  uuid = None

  def __init__(self,
               relative_path,
               namespace,
               uuid,
               disposal_timeout = 168):
    self.relative_path = relative_path
    self.namespace = namespace
    self.uuid = uuid
    self.disposal_timout = disposal_timeout

  def attributs_equal(self, other):
    if not isinstance(other, self.__class__):
      return False
    attributs = [
                "relative_path",
                "namespace",
                "uuid",
                ]
    for attr_name in attributs:
      attr = getattr(self, attr_name)
      other_attr = getattr(other, attr_name)
      if not attr == other_attr:
        return False
    return True

  def to_dict(self):
    srp_dict = {}
    attributs = [
                  "relative_path",
                  "namespace",
                  "uuid",
                  ]
    for attr_name in attributs:
      srp_dict[attr_name] = getattr(self, attr_name)
  
    return srp_dict
  
  @classmethod
  def from_dict(cls, d):
    shared_res_path = cls(relative_path="toto",
                          namespace="toto",
                          uuid="toto")
    for key, value in d.iteritems():
      setattr(shared_res_path, key, value)
    return shared_res_path



class IdGenerator(object):
  def __init__(self):
    self.current_id = 0

  def generate_id(self):
    current_id = self.current_id
    self.current_id = self.current_id + 1
    return current_id  

def to_serializable(element, 
                    id_generator,
                    transfer_ids,
                    shared_res_path_ids):
  if isinstance(element, FileTransfer):
    if element in transfer_ids:
      return transfer_ids[element]
    else:
      ident = id_generator.generate_id()
      transfer_ids[element] = ident
      return ident
  elif isinstance(element, SharedResourcePath):
    if element in shared_res_path_ids:
      return shared_res_path_ids[element]
    else:
      ident = id_generator.generate_id()
      shared_res_path_ids[element] = ident
      return ident
  elif isinstance(element, list):
    return list_to_serializable(element,
                                id_generator,
                                transfer_ids,
                                shared_res_path_ids)
  elif isinstance(element, tuple):
    return ["soma-workflow-tuple", 
            to_serializable(element[0],
            id_generator,
            transfer_ids,
            shared_res_path_ids),
            element[1]]
  else:
    return element

def from_serializable(element,
                      tr_from_ids,
                      srp_from_ids):
  if isinstance(element, list):
    if element[0] == "soma-workflow-tuple" and len(element) == 3:
      return (from_serializable(element[1], 
                                tr_from_ids, 
                                srp_from_ids),
              element[2])
   
    else:
      return list_from_serializable(element, tr_from_ids, srp_from_ids)
  elif element in tr_from_ids:
    return tr_from_ids[element]
  elif element in srp_from_ids:
    return srp_from_ids[element]
  else:
    return element

def list_to_serializable(list_to_convert,
                         id_generator,
                         transfer_ids,
                         shared_res_path_ids):
  ser_list = []
  for element in list_to_convert:
    ser_element = to_serializable(element,
                                  id_generator,
                                  transfer_ids,
                                  shared_res_path_ids)
    ser_list.append(ser_element)
  return ser_list

def list_from_serializable(list_to_convert,
                           tr_from_ids,
                           srp_from_ids):
  us_list = []
  for element in list_to_convert:
    us_element = from_serializable(element,
                                   tr_from_ids,
                                   srp_from_ids)
    us_list.append(us_element)
  return us_list






#!/usr/bin/env python

'''
@author: Soizic Laguitton

@organization: I2BM, Neurospin, Gif-sur-Yvette, France
@organization: CATI, France
@organization: U{IFR 49<http://www.ifr49.org>}

@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
'''




if __name__ == '__main__':
 
  import sys
  import logging

  import Pyro.naming
  import Pyro.core
  from Pyro.errors import PyroError, NamingError
  
  import soma_workflow.database_server
  from soma_workflow.configuration import Configuration
  
  class WorkflowDatabaseServer(Pyro.core.ObjBase, 
                               soma_workflow.database_server.WorkflowDatabaseServer):
    def __init__(self, 
                 database_file, 
                 tmp_file_dir_path):
      Pyro.core.ObjBase.__init__(self)
      soma_workflow.database_server.WorkflowDatabaseServer.__init__(self, 
                                                           database_file, 
                                                           tmp_file_dir_path)
    pass

    def test(self):
      return True
  
  if not len(sys.argv) == 2:
    sys.stdout.write("start_database_server takes 1 argument: resource id. \n")
    sys.exit(1)
  
  ressource_id = sys.argv[1]
  print "Ressource: " + ressource_id

  config = Configuration.load_from_file(ressource_id)

  (server_log_file,
   server_log_format,
   server_log_level) = config.get_server_log_info()

  if server_log_file:
    logging.basicConfig(
      filename = server_log_file,
      format = server_log_format,
      level = eval("logging."+ server_log_level))

  
  ########################
  # Pyro server creation 
  Pyro.core.initServer()
  daemon = Pyro.core.Daemon()
  # locate the NS 
  locator = Pyro.naming.NameServerLocator()
  print 'searching for Name Server...'
  name_server_host = config.get_name_server_host()
  if name_server_host == 'None':
    ns = locator.getNS()
  else: 
    ns = locator.getNS(host=name_server_host )
  daemon.useNameServer(ns)

  # connect a new object implementation (first unregister previous one)
  server_name = config.get_server_name()
  try:
    ns.unregister(server_name)
  except NamingError:
    pass

  # connect new object implementation
  server = WorkflowDatabaseServer(config.get_database_file(), 
                                  config.get_transfered_file_dir())
  daemon.connect(server, server_name)
  print "port = " + repr(daemon.port)
  
  # enter the server loop.
  print 'Server object ' + server_name + ' ready.'

  ########################
  # Request loop
  daemon.requestLoop()



'''
@author: Soizic Laguitton
@organization: U{IFR 49<http://www.ifr49.org>}
@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
'''

import unittest
import time
import os
import getpass
import sys
from datetime import datetime
from datetime import timedelta

import soma_workflow.constants as constants
from soma_workflow.client import WorkflowController, Job, FileTransfer
import soma_workflow.configuration as configuration

def checkFiles(files, filesModels, tolerance = 0):
  index = 0
  for file in files:
    t = tolerance
    (identical, msg) = identicalFiles(file, filesModels[index])
    if not identical: 
      if t <= 0: 
        return (identical, msg)
      else: 
        t = t -1
        print "\n checkFiles: "+ msg
    index = index +1
  return (True, None)


def identicalFiles(filepath1, filepath2):
  file1 = open(filepath1)
  file2 = open(filepath2)
  lineNb = 1
  line1 = file1.readline()
  line2 = file2.readline()
  identical = line1 == line2
  
  while identical and line1 :
    line1 = file1.readline()
    line2 = file2.readline()
    lineNb = lineNb + 1
    identical = line1 == line2
  
  if identical: identical = line1 == line2
  if not identical:
    return (False, "%s and %s are different. line %d: \n file1: %s file2:%s" %(filepath1, filepath2, lineNb, line1, line2))
  else:
    return (True, None)



class JobExamples(object):
  '''
  Job submission example.
  Each method submits 1 job and return the tuple (job_id, local_ouput_files, std_out_err)
  => pipeline of 4 jobs with file transfer: submitJob1, submitJob2, submitJob3, and 
  submitJob4 methods. 
  => job raising an exception with file transfer: submitExceptionJob
  => local job using user's files only (even for stdout and stderr): localCustomSubmission
  => local job regular submission: localSubmission
  '''
  
  
  
  def __init__(self, wf_ctrl, inpath, outpath, python, transfer_timeout = -24, jobs_timeout = 1):
    self.wf_ctrl = wf_ctrl
    self.inpath = inpath
    self.outpath = outpath
    self.tr_timeout = transfer_timeout
    self.jobs_timeout = jobs_timeout
    self.python = python
    
    self.job1OutputFileModels = [self.inpath + "complete/outputModels/file11",
                                 self.inpath + "complete/outputModels/file12"]
    self.job2OutputFileModels = [self.inpath + "complete/outputModels/file2"]
    self.job3OutputFileModels = [self.inpath + "complete/outputModels/file3"]
    self.job4OutputFileModels = [self.inpath + "complete/outputModels/file4"]
    
    self.job1stdouterrModels = [self.inpath + "complete/outputModels/stdoutjob1",
                                self.inpath + "complete/outputModels/stderrjob1"]
    self.job2stdouterrModels = [self.inpath + "complete/outputModels/stdoutjob2",
                                self.inpath + "complete/outputModels/stderrjob2"]
    self.job3stdouterrModels = [self.inpath + "complete/outputModels/stdoutjob3",
                                self.inpath + "complete/outputModels/stderrjob3"]
    self.job4stdouterrModels = [self.inpath + "complete/outputModels/stdoutjob4",
                                self.inpath + "complete/outputModels/stderrjob4"]
     
    self.exceptionjobstdouterr = [self.inpath + "simple/outputModels/stdout_exception_job",
                                  self.inpath + "simple/outputModels/stderr_exception_job"]
    
  def setNewConnection(self, wf_ctrl):
    '''
    For the disconnection test
    '''
    self.wf_ctrl = wf_ctrl
  
  def submitJob1(self, time=2):
    self.file11_tr = self.wf_ctrl.register_transfer(
                              FileTransfer(is_input=False,         
                                           client_path=self.outpath + "file11", 
                                           disposal_timeout=self.tr_timeout)) 

    self.file12_tr = self.wf_ctrl.register_transfer(
                              FileTransfer(is_input=False,
                                           client_path=self.outpath + "file12",        disposal_timeout=self.tr_timeout))
    
    self.file0_tr = self.wf_ctrl.register_transfer(
                              FileTransfer(True,
                                           self.inpath + "complete/" + "file0", 
                                           self.tr_timeout)) 

    script1_tr = self.wf_ctrl.register_transfer(
                              FileTransfer(True,
                                           self.inpath + "complete/" + "job1.py", 
                                           self.tr_timeout))

    stdin1_tr = self.wf_ctrl.register_transfer(
                              FileTransfer(True, 
                                           self.inpath + "complete/" + "stdin1", 
                                           self.tr_timeout))

    self.wf_ctrl.transfer_files(self.file0_tr.engine_path) 
    self.wf_ctrl.transfer_files(script1_tr.engine_path) 
    self.wf_ctrl.transfer_files(stdin1_tr.engine_path) 
    
    print "files transfered "

    job1_id = self.wf_ctrl.submit_job(Job(
                                      command=[self.python, 
                                              script1_tr, 
                                              self.file0_tr, 
                                              self.file11_tr, 
                                              self.file12_tr, 
                                              repr(time)], 
                                      referenced_input_files=[self.file0_tr, 
                                                              script1_tr, 
                                                              stdin1_tr], 
                                      referenced_output_files=[self.file11_tr, 
                                                              self.file12_tr], 
                                      stdin=stdin1_tr, 
                                      join_stderrout=False, 
                                      disposal_timeout=self.jobs_timeout, 
                                      name="job1 with transfers")) 
                                    
    return (job1_id, [self.file11_tr.engine_path, self.file12_tr.engine_path], None)
  

  def submitJob2(self, time=2):
    self.file2_tr = self.wf_ctrl.register_transfer(
                                  FileTransfer(False,
                                               self.outpath + "file2", 
                                               self.tr_timeout))

    script2_tr = self.wf_ctrl.register_transfer(
                                FileTransfer(True,
                                             self.inpath + "complete/" + "job2.py", 
                                             self.tr_timeout)) 

    stdin2_tr = self.wf_ctrl.register_transfer(
                                FileTransfer(True, 
                                             self.inpath + "complete/" + "stdin1", 
                                             self.tr_timeout))
  
    self.wf_ctrl.transfer_files(script2_tr.engine_path) 
    self.wf_ctrl.transfer_files(stdin2_tr.engine_path) 
  
    job2_id = self.wf_ctrl.submit_job(Job(command=[self.python, 
                                                script2_tr, 
                                                self.file11_tr, 
                                                self.file0_tr, 
                                                self.file2_tr, 
                                                repr(time)], 
                                        referenced_input_files=[self.file0_tr, 
                                                                self.file11_tr, 
                                                                script2_tr, 
                                                                stdin2_tr], 
                                        referenced_output_files=[self.file2_tr], 
                                        stdin=stdin2_tr, 
                                        join_stderrout=False, 
                                        disposal_timeout=self.jobs_timeout, 
                                        name="job2 with transfers"))
    return (job2_id, [self.file2_tr.engine_path], None)


  def submitJob3(self, time=2):
    self.file3_tr = self.wf_ctrl.register_transfer(
                                  FileTransfer(False,
                                               self.outpath + "file3",
                                               self.tr_timeout)) 
    
    script3_tr = self.wf_ctrl.register_transfer(
                                  FileTransfer(True, 
                                               self.inpath + "complete/" + "job3.py", 
                                               self.tr_timeout))
    stdin3_tr = self.wf_ctrl.register_transfer(
                                  FileTransfer(True,
                                               self.inpath + "complete/" + "stdin3", 
                                               self.tr_timeout))
    
    self.wf_ctrl.transfer_files(script3_tr.engine_path) 
    self.wf_ctrl.transfer_files(stdin3_tr.engine_path) 

    job3_id = self.wf_ctrl.submit_job(Job(command=[self.python, 
                                                script3_tr,  
                                                self.file12_tr,  
                                                self.file3_tr,  
                                                repr(time)], 
                                        referenced_input_files=[self.file12_tr, 
                                                                script3_tr,  
                                                                stdin3_tr], 
                                        referenced_output_files=[self.file3_tr], 
                                        stdin=stdin3_tr, 
                                        join_stderrout=False,  
                                        disposal_timeout=self.jobs_timeout, 
                                        name="job3 with transfers")) 
  
    return (job3_id, [self.file3_tr.engine_path], None)
  
  
  def submitJob4(self):
    self.file4_tr = self.wf_ctrl.register_transfer(
                                  FileTransfer(False,
                                               self.outpath + "file4",
                                               self.tr_timeout))
  
    script4_tr = self.wf_ctrl.register_transfer(
                                  FileTransfer(True,
                                               self.inpath + "complete/" + "job4.py", 
                                               self.tr_timeout)) 
    stdin4_tr = self.wf_ctrl.register_transfer(
                                  FileTransfer(True,
                                               self.inpath + "complete/" + "stdin4", 
                                               self.tr_timeout))
  
    self.wf_ctrl.transfer_files(script4_tr.engine_path) 
    self.wf_ctrl.transfer_files(stdin4_tr.engine_path) 
 
    job4_id = self.wf_ctrl.submit_job(Job( 
                                          command=[self.python, 
                                                   script4_tr, 
                                                   self.file2_tr, 
                                                   self.file3_tr, 
                                                   self.file4_tr], 
                                          referenced_input_files=[self.file2_tr, 
                                                                  self.file3_tr, 
                                                                  script4_tr, 
                                                                  stdin4_tr], 
                                          referenced_output_files=[self.file4_tr], 
                                          stdin=stdin4_tr, 
                                          join_stderrout=False, 
                                          disposal_timeout=self.jobs_timeout, 
                                          name="job4 with transfers"))
    return (job4_id, [self.file4_tr.engine_path], None)
  
  
  def submitExceptionJob(self):
    script_tr = self.wf_ctrl.register_transfer(
                                FileTransfer(True,
                                             self.inpath + "simple/exceptionJob.py",
                                             self.tr_timeout))
  
    self.wf_ctrl.transfer_files(script_tr.engine_path)
  
    job_id = self.wf_ctrl.submit_job(Job(command=[self.python, script_tr], 
                                      referenced_input_files=[script_tr], 
                                      referenced_output_files=[], 
                                      stdin=None, 
                                      join_stderrout=False, 
                                      disposal_timeout=self.jobs_timeout, 
                                      name="job with exception"))
    
    return (job_id, None, None)


  def localCustomSubmission(self):
    stdout = self.outpath + "stdout_local_custom_submission"
    stderr = self.outpath + "stderr_local_custom_submission"
    file11 = self.outpath + "file11"
    file12 = self.outpath + "file12"
    job_id = self.wf_ctrl.submit_job(Job(command = [self.python, 
                                                       self.inpath + "complete/" + "job1.py", 
                                                       self.inpath + "complete/" + "file0",
                                                       file11, 
                                                       file12, "2"], 
                                            stdin = self.inpath + "complete/" + "stdin1",
                                            join_stderrout=False,
                                            disposal_timeout = self.jobs_timeout,
                                            name = "job1 local custom submission",
                                            stdout_file=stdout,
                                            stderr_file=stderr,
                                            working_directory=self.outpath ))
      
    return (job_id, 
            [file11, file12],  
            [stdout, stderr])
              
  def localSubmission(self):
    file11 = self.outpath + "file11"
    file12 = self.outpath + "file12"
    job_id = self.wf_ctrl.submit_job(Job(command = [self.python, 
                                                      self.inpath + "complete/" + "job1.py", 
                                                      self.inpath + "complete/" + "file0",
                                                      file11, 
                                                      file12, "2"], 
                                            stdin = self.inpath + "complete/" + "stdin1",
                                            join_stderrout=False, 
                                            disposal_timeout = self.jobs_timeout,
                                            name = "job1 local submission"))
    return (job_id,
            [file11, file12],
            None)
     
  def mpiJobSubmission(self, node_num):
    
    #compilation 
    
    source_tr = self.wf_ctrl.register_transfer(
                                 FileTransfer(True, 
                                              self.inpath + "mpi/simple_mpi.c",
                                              self.tr_timeout))
    
    self.wf_ctrl.transfer_files(source_tr.engine_path)

    object_tr = self.wf_ctrl.register_transfer(
                                 FileTransfer(False,
                                              self.outpath + "simple_mpi.o",
                                              self.tr_timeout))
    #/volatile/laguitton/sge6-2u5/mpich/mpich-1.2.7/bin/
    #/opt/mpich/gnu/bin/
    
    mpibin = self.wf_ctrl.config.get(self.wf_ctrl.resource_id, 
                                     configuration.OCFG_PARALLEL_ENV_MPI_BIN)
    print "mpibin = " + mpibin
    
    print "source_tr.engine_path = " + source_tr.engine_path
    print "object_tr.engine_path = " + object_tr.engine_path
    compil1job_id = self.wf_ctrl.submit_job(Job(command=[ mpibin+"/mpicc", 
                                                             "-c", source_tr, 
                                                             "-o", object_tr ], 
                                             referenced_input_files=[source_tr],
                                             referenced_output_files=[object_tr],
                                             join_stderrout=False, 
                                             disposal_timeout=self.jobs_timeout,
                                             name="job compil1 mpi"))
    
    self.wf_ctrl.wait_job([compil1job_id])
    
    bin_tr = self.wf_ctrl.register_transfer(
                              FileTransfer(True,
                                           self.outpath + "simple_mpi",
                                           self.tr_timeout))
    print "bin_tr.engine_path= " + bin_tr.engine_path
    
    compil2job_id = self.wf_ctrl.submit_job(Job(command=[ mpibin+"/mpicc", 
                                                              "-o", bin_tr, 
                                                              object_tr ], 
                                             referenced_input_files=[object_tr],
                                             referenced_output_files=[bin_tr],
                                             join_stderrout=False, 
                                             disposal_timeout=self.jobs_timeout,
                                             name="job compil2 mpi"))
    
    self.wf_ctrl.wait_job([compil2job_id])
    self.wf_ctrl.delete_transfer(object_tr.engine_path)

    
    # mpi job submission
    script = self.wf_ctrl.register_transfer(
                              FileTransfer(True,
                                           self.inpath + "mpi/simple_mpi.sh",
                                           self.tr_timeout))
    
    self.wf_ctrl.transfer_files(script.engine_path)

    job_id = self.wf_ctrl.submit_job(Job(command=[ script, repr(node_num), bin_tr], 
                                      referenced_input_files=[script, bin_tr],
                                      join_stderrout=False, 
                                      disposal_timeout=self.jobs_timeout,
                                      name="parallel job mpi",
                                      parallel_job_info= (configuration.OCFG_PARALLEL_PC_MPI,node_num)))

    self.wf_ctrl.delete_job(compil1job_id)
    self.wf_ctrl.delete_job(compil2job_id)
    

    return (job_id, [source_tr.engine_path], None)
          
     
class JobsTest(unittest.TestCase):
  '''
  Abstract class for soma_workflow job and transfer common tests.
  '''
  @staticmethod
  def setupConnection(resource_id, 
                      login, 
                      password, 
                      job_examples_dir, 
                      output_dir):
    
    JobsTest.login = login
    JobsTest.password = password
    JobsTest.resource_id = resource_id
  
    JobsTest.wf_ctrl = WorkflowController(resource_id, 
                                          login, 
                                          password)
    
    JobsTest.transfer_timeout = -24 
    JobsTest.jobs_timeout = 1
  
    JobsTest.jobExamples = JobExamples(JobsTest.wf_ctrl, 
                              job_examples_dir, 
                              output_dir, 
                              'python',
                              JobsTest.transfer_timeout, 
                              JobsTest.jobs_timeout)   
  
    JobsTest.outpath = output_dir
                                   
  def setUp(self):
    raise Exception('JobTest is an abstract class. SetUp must be implemented in subclass')
                                   
   
  def tearDown(self):
    for jid in self.myJobs:
      JobsTest.wf_ctrl.delete_job(jid)
    remainingJobs = frozenset(JobsTest.wf_ctrl.jobs().keys())
    self.failUnless(len(remainingJobs.intersection(self.myJobs)) == 0)
     
                                   
  def test_jobs(self):
    res = set(JobsTest.wf_ctrl.jobs().keys())
    self.failUnless(res.issuperset(self.myJobs))
     
  def test_wait(self):
    JobsTest.wf_ctrl.wait_job(self.myJobs)
    for jid in self.myJobs:
      status = JobsTest.wf_ctrl.job_status(jid)
      self.failUnless(status == constants.DONE or 
                      status == constants.FAILED,
                      'Job %s status after wait: %s' %(jid, status))
                      
  def test_wait2(self):
    startTime = datetime.now()
    interval = 5
    JobsTest.wf_ctrl.wait_job(self.myJobs, interval)
    delta = datetime.now() - startTime
    if delta < timedelta(seconds=interval):
      for jid in self.myJobs:
        status = JobsTest.wf_ctrl.job_status(jid)
        self.failUnless(status == constants.DONE or 
                        status == constants.FAILED,
                        'Job %s status after wait: %s' %(self.myJobs[0], status))
    else:
      self.failUnless( abs(delta-timedelta(seconds=interval)) < timedelta(seconds=1))
    
  def test_restart(self):
    jobid = self.myJobs[len(self.myJobs)-1]
    JobsTest.wf_ctrl.kill_job(jobid)
    JobsTest.wf_ctrl.restart_job(jobid)
    status = JobsTest.wf_ctrl.job_status(jobid)
    self.failUnless(not status == constants.USER_ON_HOLD and  
                    not status == constants.USER_SYSTEM_ON_HOLD and
                    not status == constants.USER_SUSPENDED and
                    not status == constants.USER_SYSTEM_SUSPENDED,
                    'Job status after restart: %s' %status)
   
  def test_kill(self):
    jobid = self.myJobs[0]
    time.sleep(2)
    JobsTest.wf_ctrl.kill_job(jobid)
    job_termination_status = JobsTest.wf_ctrl.job_termination_status(jobid)
    exitStatus = job_termination_status[0]
    status = JobsTest.wf_ctrl.job_status(jobid)
    self.failUnless(status == constants.FAILED or status == constants.DONE, 
                    'Job status after kill: %s' %status)
    self.failUnless(exitStatus == constants.USER_KILLED or exitStatus == constants.FINISHED_REGULARLY, 
                    'Job exit status after kill: %s' %exitStatus)
                    
  def test_result(self):
    raise Exception('JobTest is an abstract class. test_result must be implemented in subclass')
 
 
 
 
class LocalCustomSubmission(JobsTest):
  '''
  Submission of a job using user's files only (even for stdout and stderr)
  '''
  def setUp(self):
    self.myJobs = []
    self.myTransfers = []
    info = JobsTest.jobExamples.localCustomSubmission()
    self.myJobs.append(info[0]) 
    self.outputFiles = info[1]
    self.stdouterrFiles = info[2]
   
  def tearDown(self):
    JobsTest.tearDown(self)
    for file in self.outputFiles:
      if os.path.isfile(file): os.remove(file)
    for file in self.stdouterrFiles:
      if os.path.isfile(file): os.remove(file)
  
  def test_result(self):
    jobid = self.myJobs[0]
    JobsTest.wf_ctrl.wait_job(self.myJobs)
    status = JobsTest.wf_ctrl.job_status(jobid)

    self.failUnless(status == constants.DONE,
                    'Job %s status after wait: %s' %(jobid, status))
    job_termination_status = JobsTest.wf_ctrl.job_termination_status(jobid)
    exitStatus = job_termination_status[0]
    self.failUnless(exitStatus == constants.FINISHED_REGULARLY, 
                    'Job %s exit status: %s' %(jobid, exitStatus))
    exitValue = job_termination_status[1]
    self.failUnless(exitValue == 0,
                    'Job %s exit value: %d' %(jobid,exitValue))
      
    # checking output files
    for file in self.outputFiles:
      self.failUnless(os.path.isfile(file), 'File %s doesn t exit' %file)
      
    (correct, msg) = checkFiles(self.outputFiles, JobsTest.jobExamples.job1OutputFileModels)
    self.failUnless(correct, msg)
      
    # checking stderr and stdout files
    for file in self.stdouterrFiles:
      self.failUnless(os.path.isfile(file), 'File %s doesn t exit' %file)
    (correct, msg) = checkFiles(self.stdouterrFiles, JobsTest.jobExamples.job1stdouterrModels)
    self.failUnless(correct, msg)
 

class LocalSubmission(JobsTest):
  '''
  Submission of a job without transfer
  '''
  def setUp(self):
    self.myJobs = []
    self.myTransfers = []
    info = JobsTest.jobExamples.localSubmission()
    self.myJobs.append(info[0]) 
    self.outputFiles = info[1]
   
  def tearDown(self):
    JobsTest.tearDown(self)
    for file in self.outputFiles:
      if os.path.isfile(file): os.remove(file)
      
  def test_result(self):
    jobid = self.myJobs[0]
    JobsTest.wf_ctrl.wait_job(self.myJobs)
    status = JobsTest.wf_ctrl.job_status(jobid)
    self.failUnless(status == constants.DONE,
                    'Job %s status after wait: %s' %(jobid, status))
    job_termination_status = JobsTest.wf_ctrl.job_termination_status(jobid)
    exitStatus = job_termination_status[0]
    self.failUnless(exitStatus == constants.FINISHED_REGULARLY, 
                    'Job %s exit status: %s' %(jobid, exitStatus))
    exitValue = job_termination_status[1]
    self.failUnless(exitValue == 0,
                    'Job exit value: %d' %exitValue)
                    
                    
    # checking output files
    for file in self.outputFiles:
      self.failUnless(os.path.isfile(file), 'File %s doesn t exit' %file)
    
    (correct, msg) = checkFiles(self.outputFiles, JobsTest.jobExamples.job1OutputFileModels)
    self.failUnless(correct, msg)
    
    # checking stdout and stderr
    stdout = JobsTest.outpath + "/stdout_local_submission"
    stderr = JobsTest.outpath + "/stderr_local_submission"
    JobsTest.wf_ctrl.retrieve_job_stdouterr(self.myJobs[0], stdout, stderr)
    self.outputFiles.append(stdout)
    self.outputFiles.append(stderr)
    
    (correct, msg) = checkFiles(self.outputFiles[2:5], JobsTest.jobExamples.job1stdouterrModels)
    self.failUnless(correct, msg)
      
class SubmissionWithTransfer(JobsTest):
  '''
  Submission of a job with transfer
  '''
  def setUp(self):
    self.myJobs = []
    self.myTransfers = []
    info = JobsTest.jobExamples.submitJob1()
    self.myJobs.append(info[0]) 
    self.outputFiles = info[1]
    self.clientFiles = []
   
  def tearDown(self):
    JobsTest.tearDown(self)
    for file in self.clientFiles:
      if os.path.isfile(file): os.remove(file)
      
  def test_result(self):
    jobid = self.myJobs[0]
    JobsTest.wf_ctrl.wait_job(self.myJobs)
    status = JobsTest.wf_ctrl.job_status(jobid)
    self.failUnless(status == constants.DONE,
                    'Job %s status after wait: %s' %(jobid, status))
    job_termination_status = JobsTest.wf_ctrl.job_termination_status(jobid)
    exitStatus = job_termination_status[0]
    self.failUnless(exitStatus == constants.FINISHED_REGULARLY, 
                    'Job %s exit status: %s' %(jobid, exitStatus))
    exitValue = job_termination_status[1]
    self.failUnless(exitValue == 0,
                    'Job exit value: %d' %exitValue)
    
    # checking output files
    for file in self.outputFiles:
      client_file = JobsTest.wf_ctrl.transfers([file])[file][0]
      self.failUnless(client_file)
      JobsTest.wf_ctrl.transfer_files(file)
      self.failUnless(os.path.isfile(client_file), 'File %s doesn t exit' %file)
      self.clientFiles.append(client_file)
   
    (correct, msg) = checkFiles(self.clientFiles, JobsTest.jobExamples.job1OutputFileModels)
    self.failUnless(correct, msg)
    
    # checking stdout and stderr
    client_stdout = JobsTest.outpath + "/stdout_submit_with_transfer"
    client_stderr = JobsTest.outpath + "/stderr_submit_with_transfer"
    JobsTest.wf_ctrl.retrieve_job_stdouterr(self.myJobs[0], client_stdout, client_stderr)
    self.clientFiles.append(client_stdout)
    self.clientFiles.append(client_stderr)
  
    (correct, msg) = checkFiles(self.clientFiles[2:5], JobsTest.jobExamples.job1stdouterrModels)
    self.failUnless(correct, msg)
    
    


class EndedJobWithTransfer(JobsTest):
  '''
  Submission of a job with transfer
  '''
  def setUp(self):
    self.myJobs = []
    self.myTransfers = []
    info = JobsTest.jobExamples.submitJob1()
    self.myJobs.append(info[0]) 
    self.outputFiles = info[1]
    self.clientFiles = []

    JobsTest.wf_ctrl.wait_job(self.myJobs)
   
  def tearDown(self):
    JobsTest.tearDown(self)
  
  def test_result(self):
    self.failUnless(True)

    
class JobPipelineWithTransfer(JobsTest):
  '''
  Submission of a job pipeline with transfer
  '''
  def setUp(self):
    self.myJobs = []
    self.myTransfers = []
    self.clientFiles = []
    self.outputFiles = []
    
    # Job1 
    
    info1 = JobsTest.jobExamples.submitJob1()
    self.myJobs.append(info1[0]) 
    self.outputFiles.extend(info1[1])

    JobsTest.wf_ctrl.wait_job(self.myJobs)
    status = JobsTest.wf_ctrl.job_status(self.myJobs[0])
    self.failUnless(status == constants.DONE,
                    'Job %s status after wait: %s' %(self.myJobs[0], status))
    job_termination_status = JobsTest.wf_ctrl.job_termination_status(self.myJobs[0])
    exitStatus = job_termination_status[0]
    self.failUnless(exitStatus == constants.FINISHED_REGULARLY, 
                    'Job %s exit status: %s' %(self.myJobs[0], exitStatus))
    exitValue = job_termination_status[1]
    self.failUnless(exitValue == 0,
                    'Job exit value: %d' %exitValue)
                    
    # Job2 & 3
    
    info2 = JobsTest.jobExamples.submitJob2()
    self.myJobs.append(info2[0]) 
    self.outputFiles.extend(info2[1])
    
    info3 = JobsTest.jobExamples.submitJob3()
    self.myJobs.append(info3[0]) 
    self.outputFiles.extend(info3[1])

    JobsTest.wf_ctrl.wait_job(self.myJobs)
    status = JobsTest.wf_ctrl.job_status(self.myJobs[1])
    self.failUnless(status == constants.DONE,
                    'Job %s status after wait: %s' %(self.myJobs[1], status))
    job_termination_status = JobsTest.wf_ctrl.job_termination_status(self.myJobs[1])
    exitStatus = job_termination_status[0]
    self.failUnless(exitStatus == constants.FINISHED_REGULARLY, 
                    'Job %s exit status: %s' %(self.myJobs[1], exitStatus))
    exitValue = job_termination_status[1]
    self.failUnless(exitValue == 0,
                    'Job exit value: %d' %exitValue)
                    
    status = JobsTest.wf_ctrl.job_status(self.myJobs[2])
    self.failUnless(status == constants.DONE,
                    'Job %s status after wait: %s' %(self.myJobs[2], status))
    job_termination_status = JobsTest.wf_ctrl.job_termination_status(self.myJobs[2])
    exitStatus = job_termination_status[0]
    self.failUnless(exitStatus == constants.FINISHED_REGULARLY, 
                    'Job %s exit status: %s' %(self.myJobs[2], exitStatus))
    exitValue = job_termination_status[1]
    self.failUnless(exitValue == 0,
                    'Job exit value: %d' %exitValue)
    
    # Job 4
    
    info4 = JobsTest.jobExamples.submitJob4()
    self.myJobs.append(info4[0]) 
    self.outputFiles.extend(info4[1])

                    
                    
   
  def tearDown(self):
    JobsTest.tearDown(self)
    for file in self.clientFiles:
      if os.path.isfile(file): os.remove(file)
      
  def test_result(self):
    jobid = self.myJobs[len(self.myJobs)-1]
    JobsTest.wf_ctrl.wait_job(self.myJobs)
    status = JobsTest.wf_ctrl.job_status(jobid)
    self.failUnless(status == constants.DONE,
                    'Job %s status after wait: %s' %(jobid, status))
    job_termination_status = JobsTest.wf_ctrl.job_termination_status(jobid)
    exitStatus = job_termination_status[0]
    self.failUnless(exitStatus == constants.FINISHED_REGULARLY, 
                    'Job %s exit status: %s' %(jobid, exitStatus))
    exitValue = job_termination_status[1]
    self.failUnless(exitValue == 0,
                    'Job exit value: %d' %exitValue)
    
    
    # checking output files
    for file in self.outputFiles:
      client_file = JobsTest.wf_ctrl.transfers([file])[file][0]
      self.failUnless(client_file)
      JobsTest.wf_ctrl.transfer_files(file)
      self.failUnless(os.path.isfile(client_file), 'File %s doesn t exit' %file)
      self.clientFiles.append(client_file)
    
    models = JobsTest.jobExamples.job1OutputFileModels + JobsTest.jobExamples.job2OutputFileModels + JobsTest.jobExamples.job3OutputFileModels + JobsTest.jobExamples.job4OutputFileModels
    (correct, msg) = checkFiles(self.clientFiles, models)
    self.failUnless(correct, msg)
    

    # checking stdout and stderr
    client_stdout = JobsTest.outpath + "/stdout_pipeline_job1"
    client_stderr = JobsTest.outpath + "/stderr_pipeline_job1"
    JobsTest.wf_ctrl.retrieve_job_stdouterr(self.myJobs[0], client_stdout, client_stderr)
    self.clientFiles.append(client_stdout)
    self.clientFiles.append(client_stderr)
    
    client_stdout = JobsTest.outpath + "/stdout_pipeline_job2"
    client_stderr = JobsTest.outpath + "/stderr_pipeline_job2"
    JobsTest.wf_ctrl.retrieve_job_stdouterr(self.myJobs[1], client_stdout, client_stderr)
    self.clientFiles.append(client_stdout)
    self.clientFiles.append(client_stderr)
  
    client_stdout = JobsTest.outpath + "/stdout_pipeline_job3"
    client_stderr = JobsTest.outpath + "/stderr_pipeline_job3"
    JobsTest.wf_ctrl.retrieve_job_stdouterr(self.myJobs[2], client_stdout, client_stderr)
    self.clientFiles.append(client_stdout)
    self.clientFiles.append(client_stderr)
    
    client_stdout = JobsTest.outpath + "/stdout_pipeline_job4"
    client_stderr = JobsTest.outpath + "/stderr_pipeline_job4"
    JobsTest.wf_ctrl.retrieve_job_stdouterr(self.myJobs[3], client_stdout, client_stderr)
    self.clientFiles.append(client_stdout)
    self.clientFiles.append(client_stderr)
   
    models = JobsTest.jobExamples.job1stdouterrModels + JobsTest.jobExamples.job2stdouterrModels + JobsTest.jobExamples.job3stdouterrModels + JobsTest.jobExamples.job4stdouterrModels
    (correct, msg) = checkFiles(self.clientFiles[5:13], models)
    self.failUnless(correct, msg)

                    
class ExceptionJobTest(JobsTest):
  '''
  Submission of a job raising an exception
  '''
  def setUp(self):
    self.myJobs = []
    self.myTransfers = []
    info = JobsTest.jobExamples.submitExceptionJob()
    self.myJobs.append(info[0])
    self.clientFiles = []
    
  def tearDown(self):
    JobsTest.tearDown(self)
    for file in self.clientFiles:
      if os.path.isfile(file): os.remove(file)
   
  def test_result(self):
    jobid = self.myJobs[0]
    JobsTest.wf_ctrl.wait_job(self.myJobs)
    status = JobsTest.wf_ctrl.job_status(jobid)
    self.failUnless(status == constants.DONE or constants.FAILED,
                    'Job %s status after wait: %s' %(jobid, status))
    job_termination_status = JobsTest.wf_ctrl.job_termination_status(jobid)
    exitStatus = job_termination_status[0]
    self.failUnless(exitStatus == constants.FINISHED_REGULARLY, 
                    'Job %s exit status: %s' %(jobid, exitStatus))
    exitValue = job_termination_status[1]
    self.failUnless(exitValue == 1,
                    'Job exit value: %d' %exitValue)
    # checking stdout and stderr
    client_stdout = JobsTest.outpath + "/stdout_exception_job"
    client_stderr = JobsTest.outpath + "/stderr_exception_job"
    JobsTest.wf_ctrl.retrieve_job_stdouterr(jobid, client_stdout, client_stderr)
    self.clientFiles.append(client_stdout)
    self.clientFiles.append(client_stderr)
    
    (identical, msg) = checkFiles(self.clientFiles, JobsTest.jobExamples.exceptionjobstdouterr,1)
    self.failUnless(identical, msg)

    
class DisconnectionTest(JobsTest):
  '''
  Submission of a job pipeline with transfer
  '''
  def setUp(self):
    self.myJobs = []
    self.myTransfers = []
    self.clientFiles = []
    self.outputFiles = []
    
    # Job1 
    
    info1 = JobsTest.jobExamples.submitJob1()
    self.myJobs.append(info1[0]) 
    self.outputFiles.extend(info1[1])

    JobsTest.wf_ctrl.wait_job(self.myJobs)
    status = JobsTest.wf_ctrl.job_status(self.myJobs[0])
    self.failUnless(status == constants.DONE,
                    'Job %s status after wait: %s' %(self.myJobs[0], status))
    job_termination_status = JobsTest.wf_ctrl.job_termination_status(self.myJobs[0])
    exitStatus = job_termination_status[0]
    self.failUnless(exitStatus == constants.FINISHED_REGULARLY, 
                    'Job %s exit status: %s' %(self.myJobs[0], exitStatus))
    exitValue = job_termination_status[1]
    self.failUnless(exitValue == 0,
                    'Job exit value: %d' %exitValue)
                    
    # Job2 & 3
    
    info2 = JobsTest.jobExamples.submitJob2(time = 60)
    self.myJobs.append(info2[0]) 
    self.outputFiles.extend(info2[1])
    
    info3 = JobsTest.jobExamples.submitJob3(time = 30)
    self.myJobs.append(info3[0]) 
    self.outputFiles.extend(info3[1])

    time.sleep(10)
    print "Disconnection...."
    JobsTest.wf_ctrl.disconnect()
    del JobsTest.wf_ctrl
    time.sleep(20)
    print ".... Reconnection"

    JobsTest.wf_ctrl = WorkflowController(JobsTest.resource_id, 
                                       JobsTest.login, 
                                       JobsTest.password)


    JobsTest.jobExamples.setNewConnection(JobsTest.wf_ctrl)
    #time.sleep(1)
   
  def tearDown(self):
    #pass
    JobsTest.tearDown(self)
    for file in self.clientFiles:
      if os.path.isfile(file): 
        print "remove " + file 
        os.remove(file)
      
  def test_result(self):

    JobsTest.wf_ctrl.wait_job(self.myJobs)
    status = JobsTest.wf_ctrl.job_status(self.myJobs[1])
    self.failUnless(status == constants.DONE,
                    'Job %s status after wait: %s' %(self.myJobs[1], status))
    job_termination_status = JobsTest.wf_ctrl.job_termination_status(self.myJobs[1])
    exitStatus = job_termination_status[0]
    self.failUnless(exitStatus == constants.FINISHED_REGULARLY, 
                    'Job %s exit status: %s' %(self.myJobs[1], exitStatus))
    exitValue = job_termination_status[1]
    self.failUnless(exitValue == 0,
                    'Job exit value: %d' %exitValue)
                    
    status = JobsTest.wf_ctrl.job_status(self.myJobs[2])
    self.failUnless(status == constants.DONE,
                    'Job %s status after wait: %s' %(self.myJobs[2], status))
    job_termination_status = JobsTest.wf_ctrl.job_termination_status(self.myJobs[2])
    exitStatus = job_termination_status[0]
    self.failUnless(exitStatus == constants.FINISHED_REGULARLY, 
                    'Job %s exit status: %s' %(self.myJobs[2], exitStatus))
    exitValue = job_termination_status[1]
    self.failUnless(exitValue == 0,
                    'Job exit value: %d' %exitValue)
    
    # Job 4
    
    info4 = JobsTest.jobExamples.submitJob4()
    self.failUnless(not info4[0] == -1, "The job was not submitted.")
    self.myJobs.append(info4[0]) 
    self.outputFiles.extend(info4[1])

    jobid = self.myJobs[len(self.myJobs)-1]
    JobsTest.wf_ctrl.wait_job(self.myJobs)
    status = JobsTest.wf_ctrl.job_status(jobid)
    self.failUnless(status == constants.DONE,
                    'Job %s status after wait: %s' %(jobid, status))
    job_termination_status = JobsTest.wf_ctrl.job_termination_status(jobid)
    exitStatus = job_termination_status[0]
    self.failUnless(exitStatus == constants.FINISHED_REGULARLY, 
                    'Job %s exit status: %s' %(jobid, exitStatus))
    exitValue = job_termination_status[1]
    self.failUnless(exitValue == 0,
                    'Job exit value: %d' %exitValue)
    
    
    # checking output files
    for file in self.outputFiles:
      client_file = JobsTest.wf_ctrl.transfers([file])[file][0]
      self.failUnless(client_file)
      JobsTest.wf_ctrl.transfer_files(file)
      self.failUnless(os.path.isfile(client_file), 'File %s doesn t exit' %file)
      self.clientFiles.append(client_file)
    
    models = JobsTest.jobExamples.job1OutputFileModels + JobsTest.jobExamples.job2OutputFileModels + JobsTest.jobExamples.job3OutputFileModels + JobsTest.jobExamples.job4OutputFileModels
    (correct, msg) = checkFiles(self.clientFiles, models)
    self.failUnless(correct, msg)
    

    # checking stdout and stderr
    client_stdout = JobsTest.outpath + "/stdout_pipeline_job1"
    client_stderr = JobsTest.outpath + "/stderr_pipeline_job1"
    JobsTest.wf_ctrl.retrieve_job_stdouterr(self.myJobs[0], client_stdout, client_stderr)
    self.clientFiles.append(client_stdout)
    self.clientFiles.append(client_stderr)
    
    client_stdout = JobsTest.outpath + "/stdout_pipeline_job2"
    client_stderr = JobsTest.outpath + "/stderr_pipeline_job2"
    JobsTest.wf_ctrl.retrieve_job_stdouterr(self.myJobs[1], client_stdout, client_stderr)
    self.clientFiles.append(client_stdout)
    self.clientFiles.append(client_stderr)
  
    client_stdout = JobsTest.outpath + "/stdout_pipeline_job3"
    client_stderr = JobsTest.outpath + "/stderr_pipeline_job3"
    JobsTest.wf_ctrl.retrieve_job_stdouterr(self.myJobs[2], client_stdout, client_stderr)
    self.clientFiles.append(client_stdout)
    self.clientFiles.append(client_stderr)
    
    client_stdout = JobsTest.outpath + "/stdout_pipeline_job4"
    client_stderr = JobsTest.outpath + "/stderr_pipeline_job4"
    JobsTest.wf_ctrl.retrieve_job_stdouterr(self.myJobs[3], client_stdout, client_stderr)
    self.clientFiles.append(client_stdout)
    self.clientFiles.append(client_stderr)
   
    models = JobsTest.jobExamples.job1stdouterrModels + JobsTest.jobExamples.job2stdouterrModels + JobsTest.jobExamples.job3stdouterrModels + JobsTest.jobExamples.job4stdouterrModels
    (correct, msg) = checkFiles(self.clientFiles[5:13], models,1)
    self.failUnless(correct, msg)
    #pass



class MPIParallelJobTest(JobsTest):
  '''
  Submission of a parallel job (MPI)
  '''
  def setUp(self):
    self.myJobs = []
    self.myTransfers = []
    self.node_num = 4
    info = JobsTest.jobExamples.mpiJobSubmission(node_num = self.node_num)
    self.myJobs.append(info[0]) 
    self.outputFiles = info[1]
   
  def tearDown(self):
    JobsTest.tearDown(self)
    #for file in self.outputFiles:
      #if os.path.isfile(file): os.remove(file)
      
  def test_result(self):
    jobid = self.myJobs[0]
    JobsTest.wf_ctrl.wait_job(self.myJobs)
    
    status = JobsTest.wf_ctrl.job_status(jobid)
    self.failUnless(status == constants.DONE,
                    'Job %s status after wait: %s' %(jobid, status))
    job_termination_status = JobsTest.wf_ctrl.job_termination_status(jobid)
    exitStatus = job_termination_status[0]
    self.failUnless(exitStatus == constants.FINISHED_REGULARLY, 
                    'Job %s exit status: %s' %(jobid, exitStatus))
    exitValue = job_termination_status[1]
    self.failUnless(exitValue == 0,
                    'Job exit value: %d' %exitValue)
                    

    print "stdout: "
    line = JobsTest.wf_ctrl.stdoutReadLine(jobid)
    process_num = 1
    while line:
      splitted_line = line.split()
      if splitted_line[0] == "Grettings":
        self.failUnless(line.rstrip() == "Grettings from process %d!" %(process_num), 
                        "stdout line:  %sinstead of  : 'Grettings from process %d!'" %(line, process_num))
        process_num = process_num +1
      line = JobsTest.wf_ctrl.stdoutReadLine(jobid)

    self.failUnless(process_num==self.node_num, 
                    "%d process(es) run instead of %d." %(process_num-1, self.node_num))

      



if __name__ == '__main__':
  
  job_examples_dir = os.environ.get("SOMA_WORKFLOW_EXAMPLES")
  output_dir = os.environ.get("SOMA_WORKFLOW_EXAMPLES_OUT")
  if not job_examples_dir or not output_dir:
    raise RuntimeError( 'The environment variables SOMA_WORKFLOW_EXAMPLES and SOMA_WORKFLOW_EXAMPLES_OUT must be set.')
     
  sys.stdout.write("----- soma-workflow tests: JOBS -------------\n")

  config_file_path = configuration.Configuration.search_config_path()
  sys.stdout.write("Configuration file: " + config_file_path)
  resource_ids = configuration.Configuration.get_configured_resources(config_file_path)
  
  # Resource
  sys.stdout.write("Configured resources:\n")
  for i in range(0, len(resource_ids)):
    sys.stdout.write("  " + repr(i) + " -> " + repr(resource_ids[i]) + "\n")
  sys.stdout.write("Select a resource number: ")
  resource_index = int(sys.stdin.readline())
  resource_id = resource_ids[resource_index]
  sys.stdout.write("Selected resource => " + repr(resource_id) + "\n")
  sys.stdout.write("---------------------------------\n")
  login = None
  password = None
  
  config = configuration.Configuration.load_from_file(resource_id, config_file_path)

  if config.get_mode() == 'remote':
    sys.stdout.write("This is a remote connection\n")
    sys.stdout.write("login:")
    login = sys.stdin.readline()
    login = login.rstrip()
    password = getpass.getpass()
  sys.stdout.write("Login => " + repr(login) + "\n")
  sys.stdout.write("---------------------------------\n")
  
  # Job type
  job_types = ["LocalCustomSubmission", "LocalSubmission", "SubmissionWithTransfer", "ExceptionJobTest", "JobPipelineWithTransfer", "DisconnectionTest", "EndedJobWithTransfer", "MPIParallelJobTest"]
  sys.stdout.write("Jobs example to test: \n")
  sys.stdout.write("all -> all \n")
  for i in range(0, len(job_types)):
    sys.stdout.write("  " + repr(i) + " -> " + repr(job_types[i]) + "\n")
  sys.stdout.write("Select one or several job type : \n")
  selected_job_type_indexes = []
  line = sys.stdin.readline()
  line = line.rstrip()
  if line == "all":
    selected_job_type = job_types
    sys.stdout.write("Selected job types: all \n")
  else:
    for strindex in line.split(" "):
      selected_job_type_indexes.append(int(strindex))
    selected_job_type = []
    sys.stdout.write("Selected job types: \n" )
    for job_type_index in selected_job_type_indexes:
      selected_job_type.append(job_types[int(job_type_index)])
      sys.stdout.write("  => " + repr(job_types[int(job_type_index)])  + "\n")

  # Test type
  sys.stdout.write("---------------------------------\n")
  test_types = ["test_result", "test_jobs", "test_wait", "test_wait2", "test_kill", "test_restart"]
  sys.stdout.write("Tests to perform: \n")
  sys.stdout.write("all -> all \n")
  for i in range(0, len(test_types)):
    sys.stdout.write("  " + repr(i) + " -> " + repr(test_types[i]) + "\n")
  sys.stdout.write("Select one or several test : \n")
  selected_test_type_indexes = []
  line = sys.stdin.readline()
  line = line.rstrip()
  if line == "all":
    selected_test_type = test_types
    sys.stdout.write("Selected test types: all \n")
  else:
    for strindex in line.split(" "):
      selected_test_type_indexes.append(int(strindex))
    selected_test_type = []
    sys.stdout.write("Selected test types: \n")
    for test_type_index in selected_test_type_indexes:
      selected_test_type.append(test_types[int(test_type_index)])
      sys.stdout.write("  => " + repr(test_types[int(test_type_index)])  + "\n")
  sys.stdout.write("---------------------------------\n")

  JobsTest.setupConnection(resource_id, 
                           login, 
                           password, 
                           job_examples_dir, 
                           output_dir)
  suite_list =  []
  tests = selected_test_type
  if "LocalCustomSubmission" in selected_job_type:
    suite_list.append(unittest.TestSuite(map(LocalCustomSubmission, tests)))
  if "LocalSubmission" in selected_job_type:
    print "tests = " + repr(tests)
    suite_list.append(unittest.TestSuite(map(LocalSubmission, tests)))
  if "SubmissionWithTransfer" in selected_job_type:
    suite_list.append(unittest.TestSuite(map(SubmissionWithTransfer, tests)))
  if "ExceptionJobTest" in selected_job_type:
    suite_list.append(unittest.TestSuite(map(ExceptionJobTest, tests)))
  if "JobPipelineWithTransfer" in selected_job_type:
    suite_list.append(unittest.TestSuite(map(JobPipelineWithTransfer, tests)))
  if "DisconnectionTest" in selected_job_type:
    suite_list.append(unittest.TestSuite(map(DisconnectionTest, tests)))
  if "EndedJobWithTransfer" in selected_job_type:
    suite_list.append(unittest.TestSuite(map(EndedJobWithTransfer, tests)))
  if "MPIParallelJobTest" in selected_job_type:
    suite_list.append(unittest.TestSuite(map(MPIParallelJobTest, tests)))

  alltests = unittest.TestSuite(suite_list)
  unittest.TextTestRunner(verbosity=2).run(alltests)
  
  sys.exit(0)




"""named tuple implementation"""

from operator import itemgetter as _itemgetter
from keyword import iskeyword as _iskeyword
import sys as _sys

def namedtuple(typename, field_names, verbose=False):
    """Returns a new subclass of tuple with named fields.

    >>> Point = namedtuple('Point', 'x y')
    >>> Point.__doc__                   # docstring for the new class
    'Point(x, y)'
    >>> p = Point(11, y=22)             # instantiate with positional args or keywords
    >>> p[0] + p[1]                     # indexable like a plain tuple
    33
    >>> x, y = p                        # unpack like a regular tuple
    >>> x, y
    (11, 22)
    >>> p.x + p.y                       # fields also accessable by name
    33
    >>> d = p._asdict()                 # convert to a dictionary
    >>> d['x']
    11
    >>> Point(**d)                      # convert from a dictionary
    Point(x=11, y=22)
    >>> p._replace(x=100)               # _replace() is like str.replace() but targets named fields
    Point(x=100, y=22)

    """

    # Parse and validate the field names.  Validation serves two purposes,
    # generating informative error messages and preventing template injection attacks.
    if isinstance(field_names, basestring):
        field_names = field_names.replace(',', ' ').split() # names separated by whitespace and/or commas
    field_names = tuple(field_names)
    for name in (typename,) + field_names:
        if not min(c.isalnum() or c=='_' for c in name):
            raise ValueError('Type names and field names can only contain alphanumeric characters and underscores: %r' % name)
        if _iskeyword(name):
            raise ValueError('Type names and field names cannot be a keyword: %r' % name)
        if name[0].isdigit():
            raise ValueError('Type names and field names cannot start with a number: %r' % name)
    seen_names = set()
    for name in field_names:
        if name.startswith('_'):
            raise ValueError('Field names cannot start with an underscore: %r' % name)
        if name in seen_names:
            raise ValueError('Encountered duplicate field name: %r' % name)
        seen_names.add(name)

    # Create and fill-in the class template
    numfields = len(field_names)
    argtxt = repr(field_names).replace("'", "")[1:-1]   # tuple repr without parens or quotes
    reprtxt = ', '.join('%s=%%r' % name for name in field_names)
    dicttxt = ', '.join('%r: t[%d]' % (name, pos) for pos, name in enumerate(field_names))
    template = '''class %(typename)s(tuple):
        '%(typename)s(%(argtxt)s)' \n
        __slots__ = () \n
        _fields = %(field_names)r \n
        def __new__(cls, %(argtxt)s):
            return tuple.__new__(cls, (%(argtxt)s)) \n
        @classmethod
        def _make(cls, iterable, new=tuple.__new__, len=len):
            'Make a new %(typename)s object from a sequence or iterable'
            result = new(cls, iterable)
            if len(result) != %(numfields)d:
                raise TypeError('Expected %(numfields)d arguments, got %%d' %% len(result))
            return result \n
        def __repr__(self):
            return '%(typename)s(%(reprtxt)s)' %% self \n
        def _asdict(t):
            'Return a new dict which maps field names to their values'
            return {%(dicttxt)s} \n
        def _replace(self, **kwds):
            'Return a new %(typename)s object replacing specified fields with new values'
            result = self._make(map(kwds.pop, %(field_names)r, self))
            if kwds:
                raise ValueError('Got unexpected field names: %%r' %% kwds.keys())
            return result \n\n''' % locals()
    for i, name in enumerate(field_names):
        template += '        %s = property(itemgetter(%d))\n' % (name, i)
    if verbose:
        print template

    # Execute the template string in a temporary namespace
    namespace = dict(itemgetter=_itemgetter)
    try:
        exec template in namespace
    except SyntaxError, e:
        raise SyntaxError(e.message + ':\n' + template)
    result = namespace[typename]

    # For pickling to work, the __module__ variable needs to be set to the frame
    # where the named tuple is created.  Bypass this step in enviroments where
    # sys._getframe is not defined (Jython for example).
    if hasattr(_sys, '_getframe'):
        result.__module__ = _sys._getframe(1).f_globals['__name__']

    return result
